/**
	$Id: ProcessJobExecutor.java 5399 2010-09-02 12:09:15Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent.executors;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import net.sf.basedb.clients.jobagent.Agent;
import net.sf.basedb.clients.jobagent.JobExecutor;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.JobAgentSettings;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.signal.EnhancedThreadSignalHandler;
import net.sf.basedb.core.signal.Signal;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalReceiver;
import net.sf.basedb.core.signal.SignalTransporter;
import net.sf.basedb.util.Values;

/**
	This is a job executor which starts a new process for each job to be
	executed. This has the advantage that a misbehaving job can't
	interfere with other jobs or the job agent. For example, a
	call to {@link java.lang.System#exit(int)} would only result in
	the job's process getting killed. Other jobs and the job agent would not
	be affected. Another advantage is that it makes it possible to apply
	a security policy to the new process for plugins that are not trusted with
	full access to the computer. See {@link PluginDefinition#isTrusted()}.
	<p>
	The disadvantage is the overhead in time and memory of starting a new
	process.
	<p>
 	<table>
 	<tr>
	 	<th>Configuration parameter</th>
	 	<th>Default value</th>
	 	<th>Description</th>
 	</tr>
 	
 	<tr>
 		<td>agent.executor.process.java</td>
 		<td></td>
 		<td>
 			Path to the java executable to use. not specified the envoronment variable 
 			<code>JAVA_HOME</code> is used if it exists. Otherwise we let the operating
 			system find the <code>java</code> command.
 		</td>
 	</tr>
 	
 	<tr>
 		<td>agent.executor.process.options</td>
 		<td>-server</td>
 		<td>
 			Extra command line options to pass to the java executable:
 			<code>java &lt;options&gt; -cp ... className ...</code>.
 		</td>
 	</tr>
	</table>
	
	@author nicklas
	@version 2.0
	@base.modified $Date: 2010-09-02 14:09:15 +0200 (Thu, 02 Sep 2010) $
*/
public class ProcessJobExecutor
	implements JobExecutor
{

	/**
		Log job agent events.
	*/
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.clients.jobagent.executors.ProcessJobExecutor");
	
	
	private String javaBin;
	private String options;
	private SignalReceiver signalReceiver;
	private SignalHandler signalHandler;
	
	public ProcessJobExecutor()
	{}
	
	/*
		From the JobRunner interface
		-------------------------------------------
	*/
	public void init(Agent agent)
	{
		javaBin = Values.getStringOrNull(agent.getProperty("agent.executor.process.java"));
		if (javaBin == null) 
		{
			javaBin = System.getenv("JAVA_HOME");
			char sep = java.io.File.separatorChar;
			if (javaBin != null) javaBin += sep + "jre" + sep + "bin" + sep + "java";
		}
		if (javaBin == null) javaBin = "java";
		options = Values.getString(agent.getProperty("agent.executor.process.options"), "-server");
		signalReceiver = agent.getSignalReceiver();
	}
	
	public void executeJob(SessionControl sc, Agent agent, Job job, JobAgentSettings settings, 
		Job.ExecutionTime usedSlot)
	{
		log.info("Executing job: " + job);

		// Generate command for new process
		List<String> cmd = new ArrayList<String>(10);
		
		// Java executable
		cmd.add(javaBin);
		
		// Other java options
		cmd.add(options);

		// Max allowed memory
		Long maxMemory = settings == null ? null : settings.getEffectiveMaxMemory();
		if (maxMemory != null)
		{
			maxMemory = maxMemory / 1024 / 1024;
			if (maxMemory >= 1) cmd.add("-Xmx" + maxMemory + "M");
		}
		
		// Class path
		String classPath = System.getProperty("java.class.path");
		cmd.add("-cp");
		cmd.add(classPath);
		
		// ThreadJobExecutor will execute the job
		cmd.add("net.sf.basedb.clients.jobagent.executors.ThreadJobExecutor");
		cmd.add("-j"); // Job id
		cmd.add(Integer.toString(job.getId()));
		if(settings != null)
		{
			cmd.add("-s"); // JobAgentSettings id
			cmd.add(Integer.toString(settings.getId()));
		}
		cmd.add("-u"); // User login
		cmd.add(agent.getLogin());
		cmd.add("-p"); // Password
		cmd.add(agent.getPassword());
		cmd.add("-t"); // Thread priority
		cmd.add(Integer.toString(agent.getThreadPriority(job.getEstimatedExecutionTime())));
		cmd.add("-x"); // Port the job agent is listening on for remote control
		cmd.add(Integer.toString(agent.getPort()));
		
		Process process = null;
		DbControl dc = null;
		try
		{
			// Update progress on job
			dc = sc.newDbControl();
			job = Job.getById(dc, job.getId());
			job.setProgress(0, "Starting new process");
			dc.commit();
			
			if (log.isDebugEnabled())
			{
				log.debug("Executing cmd: " + Values.getString(cmd, " ", true));
			}
			ProcessBuilder builder = new ProcessBuilder(cmd);
			builder.redirectErrorStream(true);
			
			try
			{
				process = builder.start();
			}
			catch (Throwable t)
			{
				log.error("Error executing job: " + job, t);
				dc = sc.newDbControl();
				job = Job.getById(dc, job.getId());
				job.doneError(t.getMessage(), Arrays.asList(t));
				dc.commit();
				return;
			}
			
			// Set up threads for reading output
			StringWriter result = new StringWriter();
			Thread t = new Thread(new StreamRedirector(
				new InputStreamReader(process.getInputStream()), result));
			t.start();
			
			try
			{
				log.info("Waiting for process to end");
				// This handler will listen for SHUTDOWN signals only
				signalHandler = new EnhancedThreadSignalHandler(Signal.SHUTDOWN);
				signalReceiver.registerSignalHandler(signalHandler);
				int exitCode = process.waitFor();
				if (exitCode != 0)
				{
					log.error("Process ended with exit code: " + exitCode);
					log.error("Process output is: " + result.toString());
					dc = sc.newDbControl();
					job = Job.getById(dc, job.getId());
					job.doneError("Job ended with exit code: " + exitCode, 
						Arrays.asList(new Exception(result.toString())));
					dc.commit();
				}
				else
				{
					log.info("Process ended");
				}
			}
			catch (InterruptedException ex)
			{
				log.info("Job was interrupted (job agent is shutting down): " + job, ex);
				// First, send SHUTDOWN/ABORT to the job if it supports it
				dc = sc.newDbControl();
				job = Job.getById(dc, job.getId());
				SignalTransporter signalTransporter = job.getSignalTransporter();
				if (signalTransporter != null)
				{
					Collection<Signal> supported = signalTransporter.getSupportedSignals();
					if (supported == null || supported.contains(Signal.SHUTDOWN))
					{
						signalTransporter.send(Signal.SHUTDOWN);
					}
					else
					{
						signalTransporter.send(Signal.ABORT);
					}
				}
				else
				{
					job.doneError("Aborted by user", Arrays.asList(ex));
					process.destroy();
				}
				dc.commit();
			}
			finally
			{
				signalReceiver.unregisterSignalHandler(signalHandler);					
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}

	public void close()
	{}
	// -------------------------------------------
	
	/**
		Used for redirecting standard output to a string.
	*/
	public static class StreamRedirector
		implements Runnable
	{
	
		private final Reader in;
		private final Writer out;
		public StreamRedirector(Reader in, Writer out)
		{
			this.in = in;
			this.out = out;
		}
		
		public void run()
		{
			BufferedReader reader = new BufferedReader(in);
			PrintWriter writer = new PrintWriter(out);
			String line = null;
			try
			{
				while ((line = reader.readLine()) != null)
				{
					writer.println(line);
				}
			}
			catch (IOException ex)
			{}
		}
	}
	
}
