/*
	$Id: RawBioAssayTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DataFileType;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.Sample;
import net.sf.basedb.core.Scan;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
This class provides methods for transferring RawBioAssay and RawBioAssayData
from base 1 to BASE 2. Base 1 files are located in /userfiles/rawdata/

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

public class RawBioAssayTransfer extends CommonItemTransfer
{
	/**
	Logger for this class
	 */
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.RawBioAssayTransfer");
	
	private ScanTransfer scanT = null;
	private ArrayTypeTransfer arrayTypeT = null;
	private UploadFileTransfer uploadFT = null;
	private SoftwareTransfer softwareT = null;
	private SampleTransfer sampleT = null;
	private Platform generic = null;
	private DataFileType rawDataFileType = null;

	RawBioAssayTransfer()
	{}
	
	/**
	Constructs a new RawBioAssayTransfer.
	
	@param userT A finished {@link UserTransfer}
	@param groupT A finished {@link GroupTransfer}
	 */
	RawBioAssayTransfer(UserTransfer userT, GroupTransfer groupT,
			ScanTransfer scanT, ArrayTypeTransfer arrayTypeT,
			UploadFileTransfer uploadFT, SoftwareTransfer softwareT, SampleTransfer sampleT)
	{
		super(userT, groupT);
		log.debug("Checking that necessary transfers have been done before proceeding with RawBioAssay");
		if (uploadFT != null && uploadFT.isDone() && scanT != null
				&& scanT.isDone() && arrayTypeT != null && arrayTypeT.isDone())
		{
			this.uploadFT = uploadFT;
			this.scanT = scanT;
			this.arrayTypeT = arrayTypeT;
			this.softwareT = softwareT;
			this.sampleT = sampleT;
		}
		else
		{
			if (!uploadFT.isDone())log.debug("Dependency: Uploaded files haven't been transfered");
			if (!scanT.isDone()) log.debug("Dependency: Scans haven't been transfered");
			if (!arrayTypeT.isDone()) log.debug("Dependency: ArrayTypes haven't been transfered");
			dependencyFailure();
		}
	}

	public void start()
	{
		log.info("Transferring RawBioAssay...");
		int count = getInt("countRawBioAssays");
		initIdMap(count);
		startProgress(count, "RawBioAssay");
		run(prepareStatementFromFile("selectRawBioAssays"), 0, getBatchSize());
		done();
		log.info("Transferring RawBioAssay...DONE");
	}

	/**
	Creates a {@link RawBioAssay} from the current position in
	<code>ResultSet</code>.
	
	@param dc {@link DbControl} used to transfer this item
	@param rs <code>ResultSet</code> containing the data for this item.
	@return <code>RawBioAssay</code>
	@throws BaseException If a new RawBioAssay cannot be created.
	@throws SQLException If current ResultSet position is invalid or number
	        of columns in the select statement is to low.
	 */
	protected RawBioAssay createItem(DbControl dc, ResultSet rs)
			throws BaseException, SQLException
	{
		if (log.isDebugEnabled()) log.debug("transfer RawBioAssay with id:" + rs.getInt(1) + " from  BASE1");
		if (generic == null) generic = Platform.getByExternalId(dc, Platform.GENERIC);
		RawBioAssay rawBioAssay = RawBioAssay.getNew(
				dc, generic, RawDataTypes.getSafeRawDataType("genepix"));
		setCommonItemData(dc, rawBioAssay, rs);
		File file = uploadFT.getRawBioAssayFileById(dc, rs.getInt(1));
		Scan scan = scanT.getScanById(dc, rs.getInt(9));
		rawBioAssay.setScan(scan);
		rawBioAssay.setSoftware(softwareT.getSoftwareById(dc, rs.getInt(11)));
		if (file != null)
		{
			if (rawDataFileType == null) 
			{
				rawDataFileType = DataFileType.getByExternalId(dc, DataFileType.GENERIC_RAW_DATA);
			}
			rawBioAssay.getFileSet().setMember(file, rawDataFileType);
		}
		// rawBioAssay.setFile(file);
		ArrayDesign ad = arrayTypeT.getArrayDesignById(dc, rs.getInt(20));
		if (ad != null)
		{
			rawBioAssay.setArrayDesign(ad);			
		}
		else
		{
			log.info("RawBioAssay " + rs.getInt(1) + " doesn't have an ArrayDesign");
		}
		String fileHeader = rs.getString(22);
		if (fileHeader != null)
		{
			rawBioAssay.setHeader("Complete Header", fileHeader);
			parseFileHeader(rawBioAssay, fileHeader);
		}
		
		// Inherit annotations from sample
		if (sampleT.isDone())
		{
			log.debug("Setting inherited annotations from sample");
			PreparedStatement ps = prepareStatementFromFile("getCh1SampleForRawBioAssay");
			ps.setInt(1, rs.getInt(1));
			ResultSet ss = ps.executeQuery();
			if (ss.next())
			{
				int base1SampleId = ss.getInt(1);
				if (base1SampleId != 0)
				{
					Sample s = sampleT.getSampleById(dc, base1SampleId);
					if (s.isAnnotated())
					{
						rawBioAssay.getAnnotationSet().inheritAnnotationSet(s.getAnnotationSet());
					}
				}
			}
			ss.close();
		}
		if (log.isDebugEnabled())
		{
			log.debug("Transfered RawBioAssay:" + rawBioAssay.getName() + "\tBASE1 id(" + rs.getInt(1) + ")");
			log.debug("--------------------------------------------------------------------------");
		}
		return rawBioAssay;
	}

	/**
	Fileheaders containing parameters on the form 'parameter=value' are set
	on the RawBioAssay.
	 */
	public void parseFileHeader(RawBioAssay rba, String fileHeader)
	{
		if (fileHeader != null && rba != null)
		{
			String regex = "\\\"([^=]+)=(.*)\\\"[\\r\\n|\\r|\\n]?";
			Pattern pattern = Pattern.compile(regex);
			Matcher matcher = pattern.matcher(fileHeader);
			while (matcher.find())
			{
				rba.setHeader(matcher.group(1), matcher.group(2));
			}
		}
	}

	/**
	@return RawBioAssay from the given base 1 id or null if no such
	        RawBioAssay has been transferrend.
	 */
	public RawBioAssay getRawBioAssayById(DbControl dc, int base1Id)
	{
		try
		{
			Integer id = getBase2Id(base1Id);
			return id == null ? null : RawBioAssay.getById(dc, id);
		}
		catch (Exception e)
		{
			log.warn("Couldn't get RawBioAssay with BASE1 id: " + base1Id);
			if (log.isDebugEnabled()) log.debug("\t" + e);
		}
		return null;
	}
}
