/*
	$Id: WizzzardTransfer.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.migrate;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginConfigurationRequest;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.PluginResponse;
import net.sf.basedb.core.plugin.Response;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
<dl>
<dt><b>Base 1 tables</b></dt>
<dd>Wizzard, Plate, RawBioAssay</dd>
<dt><b>Unused columns</b></dt>
<dd>addedDate</dd>
<dt><b>Notes</b></dd>
<dd>
<ul>
<li>A Wizzzard is a PluginConfiguration in BASE 2</li>
</ul>
</dd>
</dl>

@author Gregory, Martin
@version 2.0
@base.modified $Date: 2009-04-06 14:52:39 +0200 (Mon, 06 Apr 2009) $
*/

public class WizzzardTransfer extends CommonItemTransfer
{
	private static final org.apache.log4j.Logger log = org.apache.log4j.Logger.getLogger("net.sf.basedb.clients.migrate.WizzzardTransfer");

	/**
	@param userT A finished {@link UserTransfer}
	@param groupT A finished {@link GroupTransfer}
	 */
	WizzzardTransfer(UserTransfer userT, GroupTransfer groupT)
	{
		super(userT, groupT);
	}

	public void start()
	{
		log.info("Transferring wizzzards...");
		int count = getInt("countWizzzards");
		initIdMap(count);
		startProgress(count, "Wizzzards");
		runUnBatched(prepareStatementFromFile("selectWizzzards"), 0, getBatchSize());
		done();
	}

	public void createItem(ResultSet rs)
	{
		DbControl dc = newDbControl();
		String base1Properties = null; 
		try
		{
			base1Properties = rs.getInt(ID) + ":" + rs.getString(NAME);
			log.debug("Transferring Wizzard: " + base1Properties);
			String formatType = rs.getString(9);
			PluginDefinition pluginDefinition = getPluginDefinition(
					dc, formatType);
			PluginConfiguration pluginConfiguration = pluginDefinition.newPluginConfiguration();
			setCommonItemData(dc, pluginConfiguration, rs);
			PluginConfigurationRequest request = pluginConfiguration.configure();
			// Always allow complex column mappings, just in case
			request.setParameterValue("complexExpressions", "allow");
			
			String dataHeaderRegexp = fixRegexp(rs.getString(12));
			String dataFooterRegexp = fixRegexp(rs.getString(16));
			String fieldSeparator = fixRegexp(rs.getString(17));
			int minDataColumns = rs.getInt(10);
			int maxDataColumns = rs.getInt(11);
			
			HashMap<String, String> mappings = getParameterMappings(rs.getString(15));
			// Plugin dependent parameters
			if (formatType != null && formatType.equals("result"))
			{
				log.debug("setting parameter value");
				request.setParameterValue("rawDataType", "genepix");
				PluginResponse response = request.invoke();
				if (response.getStatus() == Response.Status.ERROR)
				{
					throw new BaseException(response.getMessage(),
							response.getErrorList().get(0));
				}
				else
				{
					// General parameters
					request = response.getNextRequest();
					request.setParameterValue("dataSplitterRegexp", fieldSeparator);
					request.setParameterValue("dataHeaderRegexp", dataHeaderRegexp);
					request.setParameterValue("dataFooterRegexp", dataFooterRegexp);
					request.setParameterValue("minDataColumns", minDataColumns);
					request.setParameterValue("maxDataColumns", maxDataColumns);
					// Always allow complex column mappings, just in case
					request.setParameterValue("complexExpressions", "allow");
					// RawBioAssayData
					request.setParameterValue(
							"blockColumnMapping", mappings.get("block"));
					request.setParameterValue(
							"columnColumnMapping", mappings.get("numCol"));
					// request.setParameterValue("metaGridXColumnMapping",
					// mappings.get("x") );
					// request.setParameterValue("metaGridYColumnMapping",
					// mappings.get("y") );
					request.setParameterValue(
							"reporterIdColumnMapping",
							mappings.get("reporterId"));
					request.setParameterValue(
							"rowColumnMapping", mappings.get("numRow"));
					request.setParameterValue(
							"xColumnMapping", mappings.get("x"));
					request.setParameterValue(
							"yColumnMapping", mappings.get("y"));
					request.setParameterValue(
							"propertyMapping.diameter", mappings.get("dia"));
					request.setParameterValue(
							"propertyMapping.ch1FgMedian",
							mappings.get("FCh1Median"));
					request.setParameterValue(
							"propertyMapping.ch1FgMean",
							mappings.get("FCh1Mean"));
					request.setParameterValue(
							"propertyMapping.ch1FgSd", mappings.get("FCh1SD"));
					request.setParameterValue(
							"propertyMapping.ch1BgMedian",
							mappings.get("BCh1Median"));
					request.setParameterValue(
							"propertyMapping.ch1BgMean",
							mappings.get("BCh1Mean"));
					request.setParameterValue(
							"propertyMapping.ch1BgSd", mappings.get("BCh1SD"));
					request.setParameterValue(
							"propertyMapping.ch1PercSd1",
							mappings.get("percCh1SD1"));
					request.setParameterValue(
							"propertyMapping.ch1PercSd2",
							mappings.get("percCh1SD2"));
					request.setParameterValue(
							"propertyMapping.ch1PercSat",
							mappings.get("percCh1Sat"));
					request.setParameterValue(
							"propertyMapping.ch2FgMedian",
							mappings.get("FCh2Median"));
					request.setParameterValue(
							"propertyMapping.ch2FgMean",
							mappings.get("FCh2Mean"));
					request.setParameterValue(
							"propertyMapping.ch2FgSd", mappings.get("FCh2SD"));
					request.setParameterValue(
							"propertyMapping.ch2BgMedian",
							mappings.get("BCh2Median"));
					request.setParameterValue(
							"propertyMapping.ch2BgMean",
							mappings.get("BCh2Mean"));
					request.setParameterValue(
							"propertyMapping.ch2BgSd", mappings.get("BCh2SD"));
					request.setParameterValue(
							"propertyMapping.ch2PercSd1",
							mappings.get("percCh2SD1"));
					request.setParameterValue(
							"propertyMapping.ch2PercSd2",
							mappings.get("percCh2SD2"));
					request.setParameterValue(
							"propertyMapping.ch2PercSat",
							mappings.get("percCh2Sat"));
					request.setParameterValue(
							"propertyMapping.ratiosSd",
							mappings.get("ratiosSD"));
					request.setParameterValue(
							"propertyMapping.rgnRatio",
							mappings.get("rgnRatio"));
					request.setParameterValue(
							"propertyMapping.rgnR2", mappings.get("rgnR2"));
					request.setParameterValue(
							"propertyMapping.fgPixels", mappings.get("FPixels"));
					request.setParameterValue(
							"propertyMapping.bgPixels", mappings.get("BPixels"));
					request.setParameterValue(
							"propertyMapping.flags", mappings.get("flags"));
				}
			}
			else if (formatType != null && formatType.equals("plate"))
			{
				log.debug("Setting parameter values for a plate-FormatType");
				request.setParameterValue("dataSplitterRegexp", fieldSeparator);
				request.setParameterValue("dataHeaderRegexp", dataHeaderRegexp);
				request.setParameterValue("dataFooterRegexp", dataFooterRegexp);
				request.setParameterValue("minDataColumns", minDataColumns);
				request.setParameterValue("maxDataColumns", maxDataColumns);
				request.setParameterValue(
						"nameColumnMapping", mappings.get("_p_name"));
				request.setParameterValue(
						"rowColumnMapping", mappings.get("_p_y"));
				request.setParameterValue(
						"columnColumnMapping", mappings.get("_p_x"));
				request.setParameterValue(
						"reporterColumnMapping", mappings.get("reporterId"));
				request.setParameterValue(
						"barcodeColumnMapping", mappings.get("barcode"));
			}
			else if (formatType != null && formatType.equals("reporter"))
			{
				log.debug("Setting parameter values for a reporter-FormatType");
				request.setParameterValue("dataSplitterRegexp", fieldSeparator);
				request.setParameterValue("dataHeaderRegexp", dataHeaderRegexp);
				request.setParameterValue("dataFooterRegexp", dataFooterRegexp);
				request.setParameterValue("minDataColumns", minDataColumns);
				request.setParameterValue("maxDataColumns", maxDataColumns);
				request.setParameterValue(
						"reporterIdColumnMapping", mappings.get("reporterId"));
				request.setParameterValue(
						"symbolColumnMapping", mappings.get("geneSymbol"));
				request.setParameterValue(
						"nameColumnMapping", mappings.get("geneName"));
				request.setParameterValue(
						"extendedColumnMapping.clusterId",
						mappings.get("clusterId"));
				request.setParameterValue(
						"extendedColumnMapping.sequence",
						mappings.get("sequence"));
				request.setParameterValue(
						"extendedColumnMapping.vector", mappings.get("vector"));
				request.setParameterValue(
						"extendedColumnMapping.tissue", mappings.get("tissue"));
				request.setParameterValue(
						"extendedColumnMapping.library",
						mappings.get("library"));
				request.setParameterValue(
						"extendedColumnMapping.accession",
						mappings.get("accession"));
				request.setParameterValue(
						"extendedColumnMapping.nid", mappings.get("nid"));
				request.setParameterValue(
						"extendedColumnMapping.chromosome",
						mappings.get("chromosome"));
				request.setParameterValue(
						"extendedColumnMapping.cytoband",
						mappings.get("cytoBand"));
				request.setParameterValue(
						"extendedColumnMapping.markers",
						mappings.get("markers"));
				request.setParameterValue(
						"extendedColumnMapping.antibiotics",
						mappings.get("antibiotics"));
				request.setParameterValue(
						"extendedColumnMapping.locusLink",
						mappings.get("locusLink"));
			}
			else if (formatType != null && formatType.equals("reportermap"))
			{
				log.debug("Setting parameter values for a reportermap-FormatType");
				request.setParameterValue("dataSplitterRegexp", fieldSeparator);
				request.setParameterValue("dataHeaderRegexp", dataHeaderRegexp);
				request.setParameterValue("dataFooterRegexp", dataFooterRegexp);
				request.setParameterValue("minDataColumns", minDataColumns);
				request.setParameterValue("maxDataColumns", maxDataColumns);
				request.setParameterValue(
						"reporterIdColumnMapping", mappings.get("reporterId"));
				request.setParameterValue(
						"blockColumnMapping", mappings.get("_r_block"));
				request.setParameterValue(
						"rowColumnMapping", mappings.get("_r_y"));
				request.setParameterValue(
						"columnColumnMapping", mappings.get("_r_x"));
			}
			else
			{
				log.warn("Unknown FormatType is " + formatType);
			}
			PluginResponse response = request.invoke();
			if (response.getStatus() == Response.Status.DONE)
			{
				response.saveParameters(dc);
				dc.saveItem(pluginConfiguration);
				if (log.isDebugEnabled())
				{
					log.debug("PluginConfiguration: " + pluginConfiguration.getName() + " is transfered");
					log.debug("--------------------------------------------------------------------------");
				}
				
			}
			else
			{
				for (Throwable err : response.getErrorList())
				{
					log.debug(err);
				}
				throw new BaseException(response.getMessage());
			}
		}
		catch (BaseException e)
		{
			log.warn("The wizzard [" + base1Properties + "] could not be transfered:" + e.getMessage() );
			if (log.isDebugEnabled()) log.debug("", e);
		}
		catch (SQLException e)
		{
			log.error("FAILED when executing SQL statement", e);
			stop();
		}
		finally
		{
			commit(dc);
		}
	}

	private HashMap<String, String> getParameterMappings(String mapping)
	{
		String regexp = "([^:]+):([^,]*),?";
		// Split on comma, group by colon: key1:value1,key2:value2
		Pattern pattern = Pattern.compile(regexp);
		Matcher matcher = pattern.matcher(mapping);
		HashMap<String, String> mappings = new HashMap<String, String>();
		while (matcher.find())
		{
			if (matcher.group(1) != null && matcher.group(2) != null)
			{
				mappings.put(matcher.group(1), matcher.group(2));
			}
		}
		return mappings;
	}

	private PluginDefinition getPluginDefinition(DbControl dc, String formatType)
			throws BaseException
	{
		String className = null;
		if (formatType.equals("plate"))
		{
			className = "net.sf.basedb.plugins.PlateFlatFileImporter";
		}
		else if (formatType.equals("reporter"))
		{
			className = "net.sf.basedb.plugins.ReporterFlatFileImporter";
		}
		else if (formatType.equals("reportermap"))
		{
			className = "net.sf.basedb.plugins.ReporterMapFlatFileImporter";
		}
		else if (formatType.equals("result"))
		{
			className = "net.sf.basedb.plugins.RawDataFlatFileImporter";
		}
		else
		{
			className = "";
		}
		PluginDefinition pluginDefinition = null;
		try
		{
			pluginDefinition = PluginDefinition.getByClassName(dc, className);
		}
		catch (Exception e)
		{
			if (log.isDebugEnabled()) log.debug(e);
		}
		if (pluginDefinition == null)
		{
			throw new BaseException("Plugin Definition is null for formatType "
					+ formatType);
		}
		return pluginDefinition;
	}
	
	private String fixRegexp(String base1Regexp)
	{
		String fixedRegexp = null;
		if (base1Regexp != null)
		{
			fixedRegexp = base1Regexp.replaceAll("\\t", "\\\\t"); // Replace tab character with \t
		}
		return fixedRegexp;
	}
	
}
