/*
	$Id: Download.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.web.servlet;

import net.sf.basedb.core.Application;
import net.sf.basedb.core.MimeType;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.InvalidPathException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.NotLoggedInException;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.Values;

import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.net.URLDecoder;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
	 @author nicklas
	 @version 2.0
	 @base.modified $Date$
*/

public final class Download
	extends HttpServlet
{

	/**
		
	*/
	private static final long serialVersionUID = 6893272087926809158L;

	private static final Pattern findId = Pattern.compile("/-([0-9a-f]+)-(/.*)$");

	private String defaultMimeType = "text/plain";
	private String forceMimeType = null;
	private boolean download = false;
	private boolean usePathInfo = false;

	public void init()
		throws ServletException
	{
		ServletConfig cfg = getServletConfig();
		forceMimeType = cfg.getInitParameter("force_mime_type");
		String temp = cfg.getInitParameter("default_mime_type");
		if (temp != null) defaultMimeType = temp;
		download = Values.getBoolean(cfg.getInitParameter("download"));
		usePathInfo = Values.getBoolean(cfg.getInitParameter("use_path_info"));
	}

	public void doGet(HttpServletRequest request, HttpServletResponse response)
		throws IOException, ServletException
	{
		/*
			Note! The request.getPathInfo() gets us the pathInfo, but it may not be
			decoded correctly if it contains national charcters.
		*/
		String contextPath = request.getContextPath();
		String servletPath = request.getServletPath();
		String pathInfo = usePathInfo ? request.getPathInfo() : 
			URLDecoder.decode(request.getRequestURI(), "UTF-8").replace(contextPath+servletPath, "");
//		System.out.println("URI=" + pathInfo+"; servletPath="+servletPath+"; contextPath="+contextPath);
		/*
			The pathInfo is expected to be like:
			/-SESSIONCONTROL_ID-/directory path/file
			We use a regexp to get the SESSIONCONTROL_ID
			and let the Path object handle the rest.
		*/
		String ID = null;
		Path path = null;
		String sPath = null;
		Matcher m = findId.matcher(pathInfo);
		if (m.find())
		{
			ID = m.group(1);
			sPath = m.group(2);
		}
		else
		{
			sPath = pathInfo;
		}
		if (sPath.startsWith("/~")) sPath = sPath.substring(1);
		try
		{
			path = new Path(sPath, Path.Type.FILE);
		}
		catch (InvalidPathException ex)
		{
			response.sendError(HttpServletResponse.SC_BAD_REQUEST,
				"Invalid path to file: "+pathInfo);
			return;
		}

		SessionControl sc = null;
		DbControl dc = null;
		InputStream in = null;
		OutputStream out = null;
		try
		{
			/*
				Get the SessionControl from the ID. If not found we set some attributes to
				be used by the NotLoggedInException handler: /exception/not_logged_in.jsp
			*/
			try
			{
				sc = Application.getSessionControl(ID, request.getRemoteAddr());
			}
			catch (ItemNotFoundException ex)
			{
				sc = Application.newSessionControl("net.sf.basedb.clients.web", request.getRemoteAddr(), null);
				ID = sc.getId();
			}
			if (sc == null || !sc.isLoggedIn())
			{
				request.setAttribute("ID", ID);
				request.setAttribute("redirect", contextPath+servletPath+"/-"+ID+"-"+path.toString());
				throw new NotLoggedInException();
			}
			File f = null;
			dc = sc.newDbControl();
			f = File.getByPath(dc, path, false);
			String mimeType = forceMimeType != null ? forceMimeType : f.getMimeType();
			if (mimeType == null)
			{
				mimeType = MimeType.getMimeType(dc, f.getName(), defaultMimeType);
			}
			dc.close();
			response.setContentType(mimeType);
			if (f.getCharacterSet() != null) response.setCharacterEncoding(f.getCharacterSet());
			if (download) response.setHeader("Content-Disposition", "attachment;filename=\""+f.getName() + "\"");
			long size = f.getSize();
			if (size >= 0 && size < Integer.MAX_VALUE) response.setContentLength((int)size);
			in = f.getDownloadStream(0);
			out = response.getOutputStream();
			FileUtil.copy(in, out);
			in.close();
			out.flush();
			out.close();
		}
		catch (Exception ex)
		{
			throw new ServletException(ex);
		}
		finally
		{
			FileUtil.close(in);
			FileUtil.close(out);
			if (dc != null) dc.close();
		}
	}

	public void doPost(HttpServletRequest request, HttpServletResponse response)
		throws IOException, ServletException
	{
		doGet(request, response);
	}
}
