/*
	$Id: ColumnDef.java 5450 2010-10-26 10:56:30Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import java.util.Arrays;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.util.Enumeration;
import net.sf.basedb.core.ItemContext;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PropertyFilter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.clients.web.Base;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.formatter.Formatter;
import net.sf.basedb.clients.web.formatter.FormatterFactory;
import net.sf.basedb.clients.web.util.HTML;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.TagSupport;


/**
	This tag defines a column in the table. It should be placed directly after the
	{@link Table &lt;tbl:table&gt;} tag.

	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:columndef
      id=...
      clazz=...
      style=...
      property=...
      sortproperty=...
      filterproperty=...
      exportproperty=...
      datatype=int|long|float|double|string|date|boolean
      collection=true|false
      annotation=true|false
      enum=...
      enumeration=...
      smartenum=true|false
      title=...
      sortable=true|false
      filterable=true|false
      exportable=true|false
      show=always|never|auto
      tooltip=...
      formatter=...
      unit=...
   &gt;
</pre>
	<p>
	<table border=1 cellspacing=0 cellpadding=2>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>yes</td>
		<td>
			The ID of the column. Should be unique among the columns in the
			table. Do not use <code>all</code> as an ID since it is has a special
			meaning. The value of this attribute
			goes directly into the standard HTML <code>id</code> attribute.
			This means that if you need a reference to the table object in your own
			JavaScript code, you may for example use the following code:
			<pre class="code">
column = document.getElementById('&lt;id&gt;');
</pre>
			The ID is also used for cross-referencing between the different tags.
			For example, <code>&lt;tbl:data columns="..."&gt;</code> and
			<code>&lt;tbl:cell column="..."&gt;</code>.
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>columnheader</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>property</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property on the item objects of this column. If the property
			isn't a property of the root item, prepend the value with a dollar sign ($)
			to indicate that the root alias should not be used when using this property
			in a query.
		</td>
	</tr>
	<tr>
		<td>sortproperty</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property to sort on when sorting on this column. If
			not specified, it takes the same value as the <code>property</code>
			value.
		</td>
	</tr>
	<tr>
		<td>filterproperty</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property to filter on when filtering on this column. If
			not specified, it takes the same value as the <code>property</code>
			value.
		</td>
	</tr>
	<tr>
		<td>exportproperty</td>
		<td>-</td>
		<td>no</td>
		<td>
			The name of the property to export on when exporting on this column. If
			not specified, it takes the same value as the <code>property</code>
			value.
		</td>
	</tr>
	<tr>
		<td>collection</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the property can contain a collection of values or only a single value.
		</td>
	</tr>

	<tr>
		<td>title</td>
		<td>-</td>
		<td>no</td>
		<td>
			The title of the column. The value is displayed in the header row.
		</td>
	</tr>

	<tr>
		<td>sortable</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the table can be sorted by this column or not. If a true value
			is specified the title will be linked to a function that resubmits the
			page with the new sort order information.
		</td>
	</tr>
	
	<tr>
		<td>filterable</td>
		<td>false</td>
		<td>no</td>
		<td>
			If the table can be filtered by this column or not.
		</td>
	</tr>
	
	<tr>
		<td>exportable</td>
		<td>no</td>
		<td>no</td>
		<td>
			If values in this column can be exported or not.
		</td>
	</tr>

	<tr>
		<td>show</td>
		<td>auto</td>
		<td>no</td>
		<td>
			Specify if this column should be visible or hidden. If the
			value is <code>always</code> the column is always shown, if the
			value is <code>never</code> the column is never shown, if the value
			is <code>auto</code> it is only shown if it appears in the
			<code>columns</code> attribute of the <code>&lt;tbl:data&gt;</code>
			tag.
		</td>
	</tr>
	
	<tr>
		<td>enumeration</td>
		<td>-</td>
		<td>no</td>
		<td>A Enumeration&lt;String, String&gt; object containing the possible
			values of this property. This is used to generate a selection list for
			the filter.
		</td>
	</tr>
	
	<tr>
		<td>smartenum</td>
		<td>false (true for annotations)</td>
		<td>no</td>
		<td>
			If the filter for an enumeration column should be displayed as
			a free text field or a selection list. Default is to use a selection
			list. Use this option only if both the hidden key and the displayed
			value of the enumeration is the same. The "smartness" of this options
			is that as the user types text, the browser will dynamically display a list
			of options matching the entered text.
		</td>
	</tr>
	
	<tr>
		<td>multiple</td>
		<td>true</td>
		<td>no</td>
		<td>
			If it should be possible to select multiple options for an 'enumeration' or
			only a single option. The default value is true.
		</td>
	</tr>
	
	<tr>
		<td>formatter</td>
		<td>-</td>
		<td>no</td>
		<td>A {@link Formatter} that formats the values in this column. If no formatter
			is specified the value is ouput as is. Values that should be formated must
			be specified in the <code>&lt;cell value="..."&gt; attribute. Any tag content
			is outputted as entered.
		</td>
	</tr>
	<tr>
		<td>unit</td>
		<td>-</td>
		<td>no</td>
		<td>
			An options {@link Unit} object that should be unit the values are
			stored with in the database. Setting this property enables
			filter expressions with units.
		</td>
	</tr>
	</table>
	<p>
	This tag has no subtags.

	@author Nicklas
	@version 2.0
*/
public class ColumnDef
	extends TagSupport
{

	/**
		
	*/
	private static final long serialVersionUID = -272408513116029365L;

	/**
		The parent &lt;tbl:table&gt; tag.
	*/
	private Table table;
	
	/**
		Optional <code>class</code> attribute
	*/
	private String clazz = "columnheader";
	
	/**
		Optional <code>style</code> attribute
	*/
	private String style = null;
	
	/**
		The name of the property on the item class.
	*/
	private String property = null;
	
	/**
		If the property is a collection or not.
	*/
	private boolean isCollection = false;

	/**
		If the property is an annotation or not.
	*/
	private boolean isAnnotation = false;

	/**
		Optional property used for sorting.
	*/
	private String sortProperty = null;

	/**
		Optional property used for filtering.
	*/
	private String filterProperty = null;
	
	/**
		Optional property used for exporting.
	*/
	private String exportProperty = null;
	
	/**
		The type of data in the property.
	*/
	private String dataType = null;
	
	/**
		The title of the column.
	*/
	private String title = null;

	/**
		If the column should be visible or not. Valid values are: always, never, auto
	*/
	private String show = "auto";
	
	/**
		If the column is sortable or not
	*/
	private boolean sortable = false;
	
	/**
		If the column is filterable or not
	*/
	private boolean filterable = false;
	
	/**
		If the column is exportable or not
	*/
	private boolean exportable = false;
	
	/**
		A tooltip with explaining text.
	*/
	private String tooltip = null;
	
	/**
		Options for enum.
	*/
	private String jsEnum = null;
	
	private Enumeration<String, String> enumeration;
	
	private boolean smartEnum;
	
	private boolean multiple = true;
	
	private transient Formatter<?> formatter;
	
	private transient Unit unit;
	
	/*
		Taglib initialization methods
	*/
	public void setClazz(String clazz)
	{
		this.clazz = clazz;
	}
	public String getClazz()
	{
		return clazz;
	}

	public void setStyle(String style)
	{
		this.style = style;
	}
	public String getStyle()
	{
		return style;
	}
	
	public void setProperty(String property)
	{
		this.property = property;
	}
	public String getProperty()
	{
		return property;
	}
	
	public void setCollection(boolean isCollection)
	{
		this.isCollection = isCollection;
	}
	public boolean isCollection()
	{
		return isCollection;
	}

	public void setAnnotation(boolean isAnnotation)
	{
		this.isAnnotation = isAnnotation;
	}
	public boolean isAnnotation()
	{
		return isAnnotation;
	}

	public void setSortproperty(String sortProperty)
	{
		this.sortProperty = sortProperty;
	}
	public String getSortproperty()
	{
		return sortProperty == null ? property : sortProperty;
	}

	public void setFilterproperty(String filterProperty)
	{
		this.filterProperty = filterProperty;
	}
	public String getFilterproperty()
	{
		return filterProperty == null ? property : filterProperty;
	}
	
	public void setExportproperty(String exportProperty)
	{
		this.exportProperty = exportProperty;
	}
	public String getExportproperty()
	{
		return exportProperty == null ? property : exportProperty;
	}

	public void setDatatype(String dataType)
	{
		this.dataType = dataType;
	}
	public String getDatatype()
	{
		return dataType;
	}
	
	public Type getValueType()
	{
		return Type.fromValue(dataType);
	}
	
	public void setTitle(String title)
	{
		this.title = title;
	}
	public String getTitle()
	{
		return title;
	}

	public void setShow(String show)
	{
		this.show = show;
	}
	public String getShow()
	{
		return show;
	}

	public void setSortable(boolean sortable)
	{
		this.sortable = sortable;
	}
	public boolean isSortable()
	{
		return sortable;
	}

	public void setFilterable(boolean filterable)
	{
		this.filterable = filterable;
	}
	public boolean isFilterable()
	{
		return filterable;
	}

	public void setExportable(boolean exportable)
	{
		this.exportable = exportable;
	}
	public boolean isExportable()
	{
		return exportable;
	}

	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return tooltip;
	}

	public void setEnum(String jsEnum)
	{
		this.jsEnum = jsEnum;
	}
	public String getEnum()
	{
		return jsEnum;
	}
	
	public void setEnumeration(Enumeration<String, String> enumeration)
	{
		this.enumeration = enumeration;
	}
	public Enumeration<String, String> getEnumeration()
	{
		return enumeration;
	}
	
	/**
		@since 2.8
	*/
	public void setMultiple(boolean multiple)
	{
		this.multiple = multiple;
	}
	
	/**
		@since 2.7
	*/
	public void setSmartenum(boolean smartEnum)
	{
		this.smartEnum = smartEnum;
	}
	
	public void setFormatter(Formatter<?> formatter)
	{
		this.formatter = formatter;
	}
	
	public Formatter<?> getFormatter()
	{
		return formatter;
	}
	
	/**
		@since 2.9
	*/
	public void setUnit(Unit unit)
	{
		this.unit = unit;
	}
	
	/**
		@since 2.9
	*/
	public Unit getUnit()
	{
		return unit;
	}
	
	/*
		Package methods
	*/

	/*
		Display methods
	*/
	public int doStartTag()
		throws JspException
	{
		if (!(getParent() instanceof Table)) throw new JspTagException("Tag <tbl:columndef> must be inside a <tbl:table> tag");
		table = (Table)getParent();
		table.addColumnDef(this);

		PropertyFilter filter = null;
		ItemContext context = null;
		boolean isPermission = "permission".equals(getId()); // requires special handling
		boolean isFilterable = (isFilterable() || isPermission) && table.getSc() != null &&  table.getItem() != null;
		if (isFilterable)
		{
			context = table.getSc().getCurrentContext(table.getItem(), table.getSubcontext());
			if (context != null && getFilterproperty() != null)
			{
				filter = context.getPropertyFilter(getFilterproperty());
			}
		}
		boolean isVisible = table.isColumnVisible(getId());
		if (isVisible)
		{
			StringBuilder sb = new StringBuilder();
			sb.append("<td");
			if (getId() != null) sb.append(" id=\"").append(getId()).append("\"");
			if (getClazz() != null) sb.append(" class=\"").append(getClazz()).append("\"");
			if (getStyle() != null) sb.append(" style=\"").append(getStyle()).append("\"");
			sb.append(">");
			if (!isSortable())
			{
				sb.append(getTitle());
			}
			else
			{
				boolean sortOnThis = table.isSortedby(getSortproperty());
				String image = sortOnThis ? table.getDirectionImage() : "";
				ItemContext.SortDirection direction = sortOnThis ? table.getReverseDirection() : table.getDirection();
				String theTooltip = getTooltip() != null ? 
					getTooltip() : "Sort by "+getTitle() + "; hold CTRL, ALT or SHIFT to sort by multiple columns";

				sb.append("<span class=\"link\" onclick=\"Table.setSortOrder(event, '").append(table.getId()).append("', '");
				sb.append(HTML.javaScriptEncode(getSortproperty())).append("', '").append(direction.name()).append("')\" title=\"");
				sb.append(theTooltip).append("\">").append(getTitle()).append("</span>").append(image);
				
				/*
				sb.append("<a href=\"javascript:Table.setSortOrder('").append(table.getId()).append("', '");
				sb.append(getSortproperty()).append("', '").append(direction.name()).append("')\" title=\"");
				sb.append(theTooltip).append("\">").append(getTitle()).append("</a>").append(image);
				*/
			}
			sb.append("</td>\n");
			table.setColumnContent(getId(), sb.toString());
			
			sb = new StringBuilder();
			sb.append("<td class=\"propertyfilter\">");
			if (isPermission)
			{
				// Special case for permission filtering
				sb.append("<select name=\"permission\"");
				sb.append(" onchange=\"Forms.submit(event);\">\n");
				sb.append("<option value=\"\">");
				Permission permission = context.getItemPermission();
				appendPermissionOption(sb, permission, Permission.READ, "Read [R]");
				appendPermissionOption(sb, permission, Permission.USE, "Use [U]");
				appendPermissionOption(sb, permission, Permission.WRITE, "Write [W]");
				appendPermissionOption(sb, permission, Permission.DELETE, "Delete [D]");
				appendPermissionOption(sb, permission, Permission.SET_OWNER, "Change owner [O]");
				appendPermissionOption(sb, permission, Permission.SET_PERMISSION, "Change permissions [P]");
				sb.append("</select>");
				
			}
			else if (isFilterable)
			{
				Type valueType = getValueType();
				Formatter<Date> dateFormatter = FormatterFactory.getDateFormatter(table.getSc());

				String inputName = "filter:" + valueType.name() + ":";
				if (filter != null && filter.isTemporary()) inputName = "tmp" + inputName;
				if (unit != null) inputName += unit.getId() + ":";
				inputName += getFilterproperty();
				if (getEnumeration() != null)
				{
					String filterValue = filter == null ? "" : HTML.encodeTags(Base.getPropertyFilterString(filter, dateFormatter));
					if (smartEnum || isAnnotation)
					{
						sb.append("<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\" class=\"smartinput\"><tr><td>");
						sb.append("<input class=\"text\" type=\"text\"");
						sb.append(" name=\"").append(inputName).append("\"");
						sb.append(" size=\"20\"");
						sb.append(" maxlength=\"").append(PropertyFilter.MAX_VALUE_LENGTH).append("\"");
						sb.append(" value=\"").append(filterValue).append("\"");
						sb.append(">");
						sb.append("</td><td>");
						sb.append("<img class=\"smartimage\" src=\"").append(table.getPage().getRoot()).append("images/pull_down.gif\"");
						sb.append(" onclick=\"SmartEnum.activateAndDisplay('");
						sb.append(table.getId()).append("','").append(inputName).append("')\">");
						sb.append("</td></tr></table>");
						StringBuilder addSmart = new StringBuilder();
						addSmart.append("SmartEnum.enableForField('").append(table.getId()).append("'");
						addSmart.append(",'").append(inputName).append("'");
						addSmart.append(",[");
						for (int i = 0; i < enumeration.size(); ++i)
						{
							if (i > 0) addSmart.append(",");
							addSmart.append("'").append(HTML.javaScriptEncode(enumeration.getKey(i))).append("'");
						}
						addSmart.append("]);\n");
						table.addScript(addSmart.toString());
					}
					else
					{
						Set<String> selected = Collections.emptySet();
						String displayName = "display:" + getFilterproperty();
						sb.append("<table border=\"0\" cellspacing=\"0\" cellpadding=\"0\" class=\"multiselect\"><tr><td>");
						sb.append("<input type=\"hidden\"");
						sb.append(" name=\"").append(inputName).append("\"");
						sb.append(" value=\"").append(filterValue).append("\"");
						sb.append(">");
						sb.append("<input type=\"text\" class=\"text\" size=\"20\"");
						sb.append(" name=\"").append(displayName).append("\"");
						sb.append(">");
						sb.append("</td><td>");
						sb.append("<img class=\"smartimage\" src=\"").append(table.getPage().getRoot()).append("images/pull_down.gif\"");
						sb.append(" onclick=\"MultiOptions.activateAndDisplay(");
						sb.append("'").append(table.getId()).append("'");
						sb.append(",'").append(displayName).append("')\">");
						sb.append("</td></tr></table>");
						StringBuilder addOptions = new StringBuilder();
						addOptions.append("MultiOptions.enableForField(");
						addOptions.append("'").append(table.getId()).append("'");
						addOptions.append(",'").append(displayName).append("'");
						addOptions.append(",'").append(inputName).append("'");
						addOptions.append(",").append(multiple ? "1" : "0");
						addOptions.append(",[");
						if (filter != null)
						{
							if (filter.getValue() != null)
							{
								selected = new HashSet<String>(Arrays.asList(filter.getValue().split("\\|")));
							}
							else
							{
								selected = Collections.singleton("");
							}
						}
						for (int i = 0; i < enumeration.size(); ++i)
						{
							String key = enumeration.getKey(i);
							if (i > 0) addOptions.append(",");
							addOptions.append("{key:'").append(HTML.javaScriptEncode(key)).append("'");
							addOptions.append(", value:'").append(HTML.javaScriptEncode(enumeration.getValue(i))).append("'");
							addOptions.append(", selected:").append(selected.contains(key)).append("}");
						}
						addOptions.append("]);\n");
						table.addScript(addOptions.toString());
						
						/*
						sb.append("<select name=\"").append(inputName).append("\"");
						sb.append(" onchange=\"Forms.submit(event);\">\n");
						sb.append("<option value=\"\">");
						Enumeration<String, String> enumeration = getEnumeration();
						for (int i = 0; i < enumeration.size(); ++i)
						{
							String selected = filterValue.equals(enumeration.getKey(i).trim()) ? " selected" : "";
							sb.append("<option value=\"").append(enumeration.getKey(i)).append("\"").append(selected);
							sb.append(">").append(enumeration.getValue(i)).append("\n");
						}
						sb.append("</select>");
						*/
					}
				}
				else if (valueType == Type.BOOLEAN)
				{
					boolean filterValue = filter == null ? false : Values.getBoolean(filter.getValue());
					sb.append("<input type=\"radio\" name=\"").append(inputName).append("\"");
					sb.append(filter == null ? "checked" : "");
					sb.append(" value=\"\" onclick=\"Forms.submit(event)\">").append("any");
					
					sb.append("<input type=\"radio\" name=\"").append(inputName).append("\"");
					sb.append(filter != null && filterValue ? "checked" : "");
					sb.append(" value=\"true\" onclick=\"Forms.submit(event)\">").append("true");
					
					sb.append("<input type=\"radio\" name=\"").append(inputName).append("\"");
					sb.append(filter != null && !filterValue ? "checked" : "");
					sb.append(" value=\"false\" onclick=\"Forms.submit(event)\">").append("false");
				}
				else
				{
					String filterValue = HTML.encodeTags(Base.getPropertyFilterString(filter, dateFormatter));
					sb.append("<input class=\"text\" type=\"text\"");
					sb.append(" name=\"").append(inputName).append("\"");
					sb.append(" size=\"20\"");
					sb.append(" maxlength=\"").append(PropertyFilter.MAX_VALUE_LENGTH).append("\"");
					sb.append(" value=\"").append(filterValue).append("\"");
					sb.append(" onkeypress=\"return Forms.submitOnEnter(event);\">");
				}
			}
			else
			{
				sb.append("&nbsp;");	
			}
			sb.append("</td>");
			table.setColumnFilter(getId(), sb.toString());
		}
		if (isFilterable && filter != null) 
		{
			table.addFilteredColumn(getId(), isVisible);
		}
		return SKIP_BODY;
	}

	private void appendPermissionOption(StringBuilder sb, Permission selected, Permission toAdd, String display)
	{
		sb.append("<option value=\"").append(toAdd.name()).append("\"");
		sb.append(selected == toAdd ? " selected" : "").append(">").append(display).append("\n");
	}
	
}


