/*
	$Id: BioMaterial.java 5535 2010-12-21 10:31:46Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.core;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.BioMaterialData;

/**
	This is the base class for for the four types of biomaterials:
	{@link BioSource}, {@link Sample}, {@link Extract}
	and {@link LabeledExtract}.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2010-12-21 11:31:46 +0100 (Tue, 21 Dec 2010) $
*/
public abstract class BioMaterial<D extends BioMaterialData>
	extends AnnotatedItem<D>
{

	/**
		Get a biomaterial item when you know the id. This method can return any
		of the subclasses of {@link BioMaterial}. Eg. {@link BioSource},
		{@link Sample}, {@link Extract} or {@link LabeledExtract}.
		
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>Sample</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have read permission to the item
		@throws BaseException If there is another error
		@since 2.17
	*/
	public static BioMaterial getById(DbControl dc, int id)
	{
		BioMaterial bm = dc.loadItemByDataClass(BioMaterial.class, BioMaterialData.class, id);
		if (bm == null) throw new ItemNotFoundException("BioMaterial[id="+id+"]");
		return bm;
	}
	
	BioMaterial(D bioMaterialData)
	{
		super(bioMaterialData);
	}

	/**
		The maximum length of the external id that can be stored in the database.
		@see #setExternalId(String)
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = BioMaterialData.MAX_EXTERNAL_ID_LENGTH;

	
	/**
		On delete action: issue an update to decrease the size of all 
		{@link BioMaterialList}:s this is a member of.
	*/
	@Override
	void onBeforeCommit(Action action)
		throws NotLoggedInException, BaseException
	{
		super.onBeforeCommit(action);
		if (action == Action.DELETE)
		{
			org.hibernate.Session session = getDbControl().getHibernateSession();
			org.hibernate.Query query = HibernateUtil.getPredefinedSQLQuery(session, 
				"UPDATE_BIOMATERIALLIST_SIZE");
			/*
				UPDATE [BioMaterialLists] 
				SET [size] = [size] + :delta
				WHERE [id] IN (
					SELECT bml.[list_id]
					FROM [BioMaterialListMembers] bml
					WHERE bml.[biomaterial_id] = :bioMaterialId
				)
			*/
			query.setInteger("delta", -1);
			query.setInteger("bioMaterialId", getId());
			HibernateUtil.executeUpdate(query);
		}
	}
	/**
		Get the external id of the biomaterial. This value can be used to link
		with information in external databases. It is not used by the BASE core
		and it doesn't have to be unique.
		@return The external id for the biomaterial
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	/**
		Set the external id of the biomaterial. The value
		may be null but must not be longer than the value specified by
		the {@link #MAX_EXTERNAL_ID_LENGTH} constant.
		@param externalId The new value for the external id
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the string is too long
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}

	
	/**
		Count the number of already existing children to this biomaterial.
		The count includes items that the logged in user doesn't have access to.
		@param pooled If TRUE, the count is for biomaterials of the same type,
			if FALSE the count is for biomaterials of the child type
		@since 2.16
	*/
	public abstract long countChildren(boolean pooled);
	
}



