/**
	$Id: BioPlateType.java 5456 2010-10-28 12:12:26Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Set;

import net.sf.basedb.core.data.BioPlateTypeData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	This class represents the type of a bio-plate. This information
	can be used together with event definitions to create a 
	work path in the lab. Eg. start with a plate of some type T1,
	then apply event E1 to it to get a new plate with type T2,
	repeat with event E2 to get a plate with type T3, and so on
	until the biomaterial has been hybridized.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2010-10-28 14:12:26 +0200 (Thu, 28 Oct 2010) $
*/
public class BioPlateType
	extends BasicItem<BioPlateTypeData>
	implements Nameable, Removable
{

	/**
		The type of item represented by this class.
		@see Item#BIOPLATETYPE
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOPLATETYPE;

	/**
	 	Creates a new bioplate type.
	 	 
	  	@param dc DbControl used for database access
	  	@return A new BioPlate object
	  	@throws BaseException If anything goes wrong when creating
	*/
	public static BioPlateType getNew(DbControl dc)
		throws BaseException, InvalidDataException
	{
		BioPlateType bp = dc.newItem(BioPlateType.class);
		bp.setName("New bioplate type");		
		return bp;
	}

	/**
	 	Gets a BioPlate type when knowing the item's id
	  	@param dc DbControl used for database access
	  	@param id The item's id.
	  	@return <code>BioPlateType</code> object
	  	@throws ItemNotFoundException If no item with the given id was found
	  	@throws PermissionDeniedException If logged in user doesn't have read 
	  		permission to the requested item
	  	@throws BaseException If there is another error.
	*/
	public static BioPlateType getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioPlateType p = dc.loadItem(BioPlateType.class, id);
		if (p == null) throw new ItemNotFoundException("BioPlateType[id="+id+"]");
		return p;
	}
	
	/**
	 	Gets query that selects all available bioplates
	  	@return <code>ItemQuery</code> object
	 */
	public static ItemQuery<BioPlateType> getQuery()
	{
		return new ItemQuery<BioPlateType>(BioPlateType.class);
	}
	
	BioPlateType(BioPlateTypeData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}	
	//-------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------
	/*
	 	From the BasicItem class
	 	--------------------------------------
	 */
	/**
	 	Check if there are any <code>BioPlates</code> or 
	 	<code>BioPlateEventDefinitions</code> using this type.
	*/
	@Override
	public boolean isUsed()
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOPLATES_FOR_BIOPLATETYPE", "count(*)");
		/*
		 	SELECT {1}
		 	FROM BioPlateData bpl
		 	WHERE bpl.bioPlateType = :bioPlateType
		 */
		query.setEntity("bioPlateType", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		return used || super.isUsed();
	}
	
	/**
	 	Gets all <code>BioPlates</code> and <code>BioPlateEventDefinitions</code> 
	 	using this type
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		// Bioplates
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOPLATES_FOR_BIOPLATETYPE", "bpl.id");
		/*
		 	SELECT {1}
		 	FROM BioPlateData bpl
		 	WHERE bpl.bioPlateType = :bioPlateType
		 */
		query.setEntity("bioPlateType", this.getData());
		addUsingItems(using, Item.BIOPLATE, query);
		return using;
	}
	//-------------------------------------------

	
	/**
		Get the biomaterial type that can be stored on plates of
		this type.
		@return One of {@link Item#SAMPLE}, {@link Item#EXTRACT}, {@link Item#LABELEDEXTRACT}
			or null if the plate can store mixed types
	 */
	public Item getBioMaterialType()
	{
		Integer code = getData().getBioMaterialType();
		return code == null ? null : Item.fromValue(code);
	}
	
	/**
		Set the type of biomaterial that can be stored on plates with
		this type. Valid values are {@link Item#SAMPLE} {@link Item#EXTRACT}, 
		{@link Item#LABELEDEXTRACT} and null. The value can't be changed 
		after the plate type has been saved to the database.
		@param type A biomaterial item type or null
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission or if the plate type has already been saved
			to the database
		@throws InvalidDataException If the type is not a biomaterial type
	*/
	public void setBioMaterialType(Item type)
	{
		checkPermission(Permission.WRITE);
		if (isInDatabase())
		{
			throw new PermissionDeniedException(Permission.WRITE, 
				"bioMaterialType of BioPlateType[id=" + getId() + "]");
		}
		boolean ok = type == null || type == Item.SAMPLE || type == Item.EXTRACT || 
			type == Item.LABELEDEXTRACT;
		if (!ok)
		{
			throw new InvalidDataException("type must be one of SAMPLE, EXTRACT, LABELEDEXTRACT or null");
		}
		getData().setBioMaterialType(type == null ? null : type.getValue());		
	}
	
	/**
		Get the lock mode for wells located on plates of this plate type.
		The lock mode determines if biomaterial may be changed in the
		well.
		
		@return A lock mode
	*/
	public BioWell.LockMode getLockMode()
	{
		return BioWell.LockMode.fromValue(getData().getLockMode());
	}
	/**
		Set the lock mode for wells on plates of this plate type.
		@param lockMode A lock mode, null is not allowed
	*/
	public void setLockMode(BioWell.LockMode lockMode)
	{
		checkPermission(Permission.WRITE);
		if (lockMode == null) throw new InvalidUseOfNullException("lockMode");
		getData().setLockMode(lockMode.getValue());
	}
	
	/**
		Creates a query that returns all bioplates that are using
		this type.
		@return An ItemQuery object
	*/
	public ItemQuery<BioPlate> getBioPlates()
	{
		ItemQuery<BioPlate> query = BioPlate.getQuery();
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property("bioPlateType"), 
					Hql.entity(this)
				)
			);
		return query;
	}

	
}
