/*
	$Id: Config.java 5570 2011-02-16 13:49:40Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Arrays;
import java.util.Collections;
import java.util.Locale;
import java.util.Properties;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.Pattern;
import java.util.regex.PatternSyntaxException;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.net.URL;
import java.nio.charset.Charset;

import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.collections.Predicate;

import net.sf.basedb.util.ComparableComparator;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.PriorityComparator;

/**
	Use this class to access the configuration settings for BASE.
	<p>
	The configuration settings are always stored in a file named 
	<code>base.config</code>. It should be located at the root of 
	one of the paths in the CLASSPATH environment variable. For Tomcat this 
	means it should be located in the <code>WEB-INF/classes</code> folder.

	@author Nicklas
	@version 2.0
*/
public class Config
{
	/**
		Variable for holding the configuration properties.
	*/
	private static Properties config = null;

	/**
		Properties that has been programmatically overridden.
	*/
	private static Properties overridden = null;
	
	private static boolean isInitialised = false;
	
	private static java.io.File configDir = null;
	
	private static Locale locale = Locale.getDefault();
	
	private static SortedSet<String> commonCharsets;
	
	private static SortedSet<String> allCharsets;
	
	/**
		Loads the settings from the configuration file.
		@throws ConfigurationException This exception is thrown if the
			configuration settings cannot be loaded
	*/
	static synchronized void init()
		throws ConfigurationException
	{
		if (isInitialised) return;
		InputStream is = null;
		try
		{
			URL baseConfig = Config.class.getResource("/base.config");
			if (baseConfig == null)
			{
				throw new FileNotFoundException("/base.config");
			}
			configDir = new java.io.File(baseConfig.toURI()).getParentFile();
			is = baseConfig.openStream();
			config = new Properties();
			config.load(is);
			if (overridden != null) config.putAll(overridden);
			
		}
		catch (Exception ex)
		{
			config = null;
			throw new ConfigurationException("Error while loading properties from '/base.config'. ", ex);
		}
		finally
		{
			if (is != null) FileUtil.close(is);
		}

		isInitialised = true;
	
		// Load locale
		String language = getString("locale.language", "");
		String country = getString("locale.country", "");
		String variant = getString("locale.variant", "");
		if (language.length() > 0 || country.length() > 0 || variant.length() > 0)
		{
			locale = new Locale(language, country, variant);
		}
		else
		{
			locale = Locale.getDefault();
		}
		
		// Load character sets
		String charsets = getString("commonCharsets", "UTF-8,US-ASCII,ISO-8859-1").trim();
		commonCharsets = Collections.unmodifiableSortedSet(new TreeSet<String>(Arrays.asList(charsets.split("\\s*,\\s*"))));

		SortedSet<String> tmp = new TreeSet<String>(
				new PriorityComparator<String>(commonCharsets, 
				new ComparableComparator<String>()));
		tmp.addAll(Charset.availableCharsets().keySet());
		String filter = getString("ignoreCharsets", null);
		if (filter != null)
		{
			try
			{
				final Pattern p = Pattern.compile(filter);
				CollectionUtils.filter(tmp, new Predicate()
				{
					@Override
					public boolean evaluate(Object o)
					{
						return !p.matcher((String)o).matches() || commonCharsets.contains(o);
					}
				});
			}
			catch (PatternSyntaxException ex)
			{
				isInitialised = false;
				throw new ConfigurationException("Invalid pattern for 'ignoreCharsets' property in '/base.config'. ", ex);
			}
		}
		allCharsets = Collections.unmodifiableSortedSet(tmp);
	}

	/**
		Unload all settings.
	*/
	static synchronized void unload()
	{
		isInitialised = false;
		config = null;
		overridden = null;
		configDir = null;
	}
	
	/**
		Programatically set a configuration property that overrides a configuration
		setting in the base.config file. This method should be called before the 
		{@link #init()} method is called (eg. before {@link Application#start()}
		is called). The configuration can't be changed once BASE has been started.
		Properties that has been overridden are automatically cleared when BASE is
		stopped ({@link Application#stop()}, {@link #unload()}).
		@param key The name of the configuration property
		@param value The value or null to remove the overridden property
		@since 2.16
	*/
	public static synchronized void setProperty(String key, String value)
	{
		if (isInitialised) return;
		if (overridden == null) overridden = new Properties();
		if (value == null)
		{
			overridden.remove(key);
		}
		else
		{
			overridden.setProperty(key, value);
		}
	}
	
	/**
		Get the directory where the BASE configuration files are located.
		@return The directory as a <code>java.io.File</code> object
		@since 2.6
	*/
	public static java.io.File getConfigDirectory()
	{
		return configDir;
	}
	
	/**
		Get the configuration setting specified by <code>key</code> as a String value.
		@param key The key for the setting
		@return The value specified by <code>key</code> or null if it is not found
	*/
	public static String getString(String key)
	{
		return getString(key, null);
	}

	/**
		Get the configuration setting specified by <code>key</code> as a String value.
		@param key The key for the setting
		@param defaultValue The value to return if the <code>key</code> is not found
		@return The value specified by <code>key</code> or <code>defaultValue</code> 
			if it is not found or is empty
	*/
	public static String getString(String key, String defaultValue)
	{
		if (!isInitialised) init();
		String value = config.getProperty(key, defaultValue);
		if (value != null) value = value.trim();
		return "".equals(value) ? defaultValue : value;
	}

	/**
		Get the configuration setting specified by <code>key</code> as an integer value.
		@param key The key for the setting
		@return The value specified by <code>key</code> or 0 if it is not found or
			is not a number
	*/
	public static int getInt(String key)
	{
		return getInt(key, 0);
	}

	/**
		Get the configuration setting specified by <code>key</code> as an integer value.

		@param key The key for the setting
		@param defaultValue The value to return if the <code>key</code> is not found
		@return The value specified by <code>key</code> or <code>defaultValue</code> 
			if it is not found
	*/
	public static int getInt(String key, int defaultValue)
	{
		if (!isInitialised) init();
		int result = defaultValue;
		String setting = config.getProperty(key);
		if (setting != null)
		{
			try
			{
				result = Integer.parseInt(setting);
			}
			catch (NumberFormatException ex)
			{}
		}
		return result;
	}
	
	/**
		Get the configuration setting specified by <code>key</code> as a long value.
	
		@param key The key for the setting
		@param defaultValue The value to return if the <code>key</code> is not found
		@return The value specified by <code>key</code> or <code>defaultValue</code> 
			if it is not found
	*/
	public static long getLong(String key, long defaultValue)
	{
		if (!isInitialised) init();
		long result = defaultValue;
		String setting = config.getProperty(key);
		if (setting != null)
		{
			try
			{
				result = Long.parseLong(setting);
			}
			catch (NumberFormatException ex)
			{}
		}
		return result;
	}

	/**
		Get the configuration setting specified by <code>key</code> as a boolean 
		value. FALSE is returned if the setting doesn't exist, if it contains an
		empty string, or one of the values 'no', 'false' or '0'.
		In all other cases TRUE is returned.
		@param key The key for the setting
		@return The value specified by <code>key</code>
	*/
	public static boolean getBoolean(String key)
	{
		if (!isInitialised) init();
		String setting = config.getProperty(key);
		return (setting == null
			|| setting.equals("")
			|| setting.equals("0")
			|| setting.equals("no")
			|| setting.equals("false")
			) ? false : true;
	}
	
	/**
		Get the default character set to use when no other has been 
		specified. It reads the setting from the <code>defaultCharset</code>
		setting and uses <code>ISO-8859-1</code> if not specified.
		@return The name of the default charset to use
		@since 2.1.1
	*/
	public static String getCharset()
	{
		return getString("defaultCharset", "ISO-8859-1");
	}

	/**
		Get a set containing the most common character sets used by text files.
		The set is sorted by name.
		@return A sorted set
		@since 2.17
	*/
	public static SortedSet<String> getCommonCharsets()
	{
		return commonCharsets;
	}
	
	/**
		Get a set containing all character sets known to the current java virtual machine.
		The set is sorted by name with the most common character sets first.
		@return A sorted set
		@since 2.17
	*/
	public static SortedSet<String> getAllCharsets()
	{
		return allCharsets;
	}
	
	/**
		Get the default locale configured for the server.
		@since 2.16
	*/
	public static Locale getLocale()
	{
		return locale;
	}

}
