/*
  $Id: Hardware.java 5452 2010-10-27 12:32:44Z nicklas $

	Copyright (C) 2005 Johan Enell, Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

  BASE is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 3
  of the License, or (at your option) any later version.

  BASE is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.data.HardwareData;

/**
	This class is used to represent individual hardware items
	and information about them. In the current version of BASE 
	the only type of hardware we keep information about is scanners.
	
	@author enell
	@version 2.0
	@see HardwareType HardwareType
	@base.modified $Date: 2010-10-27 14:32:44 +0200 (Wed, 27 Oct 2010) $
 */
public class Hardware
	extends CommonItem<HardwareData>
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#HARDWARE
		@see #getType()
	*/
	public static final Item TYPE = Item.HARDWARE;

	/**
		The maximum length of the version variable that can be stored in the
		database. Check the length against this value before calling the
		{@link #setVersionString(String)} method to avoid
		exceptions.
	*/
	public static final int MAX_VERSIONSTRING_LENGTH = HardwareData.MAX_VERSIONSTRING_LENGTH;

	/**
		Create a new <code>Hardware</code> item.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param hardwareType The {@link HardwareType} connected to this hardware. It
			cannot be null.
		@return The new <code>Hardware</code> item
		@throws BaseException If there is an error
	*/
	public static Hardware getNew(DbControl dc, HardwareType hardwareType)
		throws BaseException
	{
		Hardware hardware = dc.newItem(Hardware.class);
		hardware.setName("New hardware");
		hardware.setHardwareType(hardwareType);
		hardware.getData().setEntryDate(new Date());
		return hardware;
	}

	/**
		Get a <code>Hardware</code> item when you know the ID.

		@param dc The {@link DbControl} which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Hardware</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ } permission to the item
		@throws BaseException If there is another error
	*/
	public static Hardware getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Hardware hw = dc.loadItem(Hardware.class, id);
		if (hw == null) throw new ItemNotFoundException("Hardware[ID="+id+"]");
		return hw;
	}
	
	/**
		Get a query configured to retrieve hardware.
		
		@return An {@link ItemQuery} object
	 	@throws BaseException When the query could not be configured. 
	*/
	public static ItemQuery<Hardware> getQuery()
		throws BaseException
	{
		return new ItemQuery<Hardware>(Hardware.class);
	}

	
	/**
		Creates a new hardware item.
		
		@param data the data
	*/
	Hardware(HardwareData data)
	{
		super(data);
	}
	
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>a {@link Scan} is linked to this hardware
		<li>an {@link ArrayBatch} is linked to this hardware
		<li>a {@link BioMaterialEvent} is linked to this hardware
		<li>a {@link PlateEvent} is linked to this hardware
		</ul>
		
		@return TRUE if this hardware is used.
		@throws BaseException if there is any error.
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		boolean used = false;
		if (!used)
		{
			org.hibernate.Query q = HibernateUtil.getPredefinedQuery(session, 
				"GET_SCANS_FOR_SCANNER", "count(*)");
			/*
				SELECT {1}
				FROM ScanData sc
				WHERE sc.scanner = :scanner
			*/
			q.setEntity("scanner", this.getData());
			used = HibernateUtil.loadData(Long.class, q) > 0;
		}
		if (!used)
		{
			org.hibernate.Query q = HibernateUtil.getPredefinedQuery(session, 
				"GET_ARRAYBATCHES_FOR_PRINTROBOT", "count(*)");
			/*
				SELECT {1}
				FROM ArrayBatchData ab
				WHERE ab.printRobot = :printrobot
			*/
			q.setEntity("printrobot", this.getData());
			used = HibernateUtil.loadData(Long.class, q) > 0;
		}
		if (!used)
		{
			org.hibernate.Query q = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOMATERIALEVENTS_FOR_HARDWARE", "count(*)");
				/*
					SELECT {1}
					FROM BioMaterialEventData bme
					WHERE bme.hardware = :hardware
				*/
			q.setEntity("hardware", this.getData());
			used = HibernateUtil.loadData(Long.class, q) > 0;
		}
		if (!used)
		{
			org.hibernate.Query q = HibernateUtil.getPredefinedQuery(session, 
				"GET_PLATEEVENTS_FOR_HARDWARE", "count(*)");
				/*
					SELECT {1}
					FROM PlateEventData ple
					WHERE ple.hardware = :hardware
				*/
			q.setEntity("hardware", this.getData());
			used = HibernateUtil.loadData(Long.class, q) > 0;
		}
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Scan}:s with this hardware
		<li>{@link ArrayBatch}:s with this hardware
		<li>{@link BioMaterial}:s and {@link Hybridization}:s with this hardware
		<li>{@link Plate}:s with this hardware
		<ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		
		// Scans
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_SCANS_FOR_SCANNER", "scn.id");
			/*
				SELECT {1}
				FROM ScanData scn
				WHERE sc.scanner = :scanner
			*/
		query.setEntity("scanner", this.getData());
		addUsingItems(using, Item.SCAN, query);

		// Array batches
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_ARRAYBATCHES_FOR_PRINTROBOT", "ab.id");
			/*
				SELECT {1}
				FROM ArrayBatchData ab
				WHERE ab.printRobot = :printrobot
			*/
		query.setEntity("printrobot", this.getData());
		addUsingItems(using, Item.ARRAYBATCH, query);
		
		// BioMaterial events - biomaterial
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_BIOMATERIALEVENTS_FOR_HARDWARE", "bme.bioMaterial");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				WHERE bme.hardware = :hardware
			*/
		query.setEntity("hardware", this.getData());
		addUsingItems(using, query);
		
		// BioMaterial events - hybridizations
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_BIOMATERIALEVENTS_FOR_HARDWARE", "bme.hybridization");
			/*
				SELECT {1}
				FROM BioMaterialEventData bme
				WHERE bme.hardware = :hardware
			*/
		query.setEntity("hardware", this.getData());
		addUsingItems(using, query);

		// Plates
		query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATEEVENTS_FOR_HARDWARE", "ple.plate.id");
			/*
				SELECT {1}
				FROM PlateEventData ple
				WHERE ple.hardware = :hardware
			*/
		query.setEntity("hardware", this.getData());
		addUsingItems(using, Item.PLATE, query);
		return using;
	}
	// -------------------------------------------

	/**
		Get the associated {@link HardwareType} item.
		All <code>Hardware</code> items must have a type.
		@return The <code>HardwareType</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ } permission to the item
		@throws BaseException If there is any error
	*/
	public HardwareType getHardwareType()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(HardwareType.class, getData().getHardwareType());
	}

	/**
		Set the {@link HardwareType} of this <code>Hardware</code> item.
		This parameter mustn't be <code>null</code>.
		@param hardwareType The new <code>HardwareType</code>
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#WRITE} permission for the hardware or
			{@link Permission#USE} permission for the hardware type
		@throws InvalidUseOfNullException If the hardware type is null
	*/
	public void setHardwareType(HardwareType hardwareType)
		throws PermissionDeniedException, InvalidUseOfNullException
	{
		checkPermission(Permission.WRITE);
		if (hardwareType == null) throw new InvalidUseOfNullException("hardwareType");
		hardwareType.checkPermission(Permission.USE);
		getData().setHardwareType(hardwareType.getData());
	}
	
	/**
		Get the versionstring of this <code>Hardware</code> item.
		@return A string with the version of this item, or null if not known
	*/
	public String getVersionString()
	{
		return getData().getVersionString();
	}

	/**
		Set the versionstring for this <code>Hardware</code> item. The value
		must not be longer than the value specified by the {@link #MAX_VERSIONSTRING_LENGTH} constant.
		
		@param versionString The new version for this item, or null if not known
		@throws StringTooLongException If versionString is longer then {@link #MAX_VERSIONSTRING_LENGTH}
		@throws PermissionDeniedException If the logged in user don't have write permission on this item
	*/
	public void setVersionString(String versionString) 
		throws StringTooLongException, PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setVersionString(StringUtil.setNullableString(versionString, "versionString", MAX_VERSIONSTRING_LENGTH));
	}
}
