/*
	$Id: Message.java 5453 2010-10-28 06:20:24Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.MessageData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.util.EmailUtil;

import java.util.Date;

import org.apache.commons.mail.Email;
import org.apache.commons.mail.EmailException;


/**
	This class represents a message. Messages are normally sent by the system
	when a job is completed. It is also possible to send messages between
	other users.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2010-10-28 08:20:24 +0200 (Thu, 28 Oct 2010) $
*/
public class Message
	extends BasicItem<MessageData>
	implements Nameable, Removable
{
	/**
		The type of item represented by this class.
		@see Item#MESSAGE
		@see #getType()
	*/
	public static final Item TYPE = Item.MESSAGE;

	/**
		Create a new <code>Message</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
	 	@param to The user who should receive the message. 
	 	@param fromName Name of the sender.
	 	@param fromUser The user who sends this message.
	 	@param job Which job this message is about.
		@return The new <code>Message</code> item
		@throws BaseException If there is an error
	*/
	public static Message getNew(DbControl dc, User to, String fromName, User fromUser, Job job)
		throws BaseException
	{
		Message m = dc.newItem(Message.class);
		m.setName("New message");
		m.setTo(to);
		m.setFrom(fromName);
		m.setSender(fromUser);
		m.getData().setTimeSent(new Date());
		m.setJob(job);
		return m;
	}

	/**
		Create a new <code>Message</code> from the system. This will not fail due
		to missing create permission for the logged in user.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
	 	@param to The user who should receive the message. 
	 	@param fromName Name of the sender.
	 	@param fromUser The user who sends this message.
	 	@param job Which job this message is about.
		@return The new <code>Message</code> item
		@throws BaseException If there is an error
	*/
	static Message getNewSystemMessage(DbControl dc, User to, String fromName, User fromUser, Job job)
	{
		Message m = dc.newItem(Message.class);
		m.systemCreated = true;
		m.setName("New message");
		m.setTo(to);
		m.setFrom(fromName);
		m.setSender(fromUser);
		m.getData().setTimeSent(new Date());
		m.setJob(job);
		return m;
	}
	
	/**
		Get a <code>Message</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param id The id of the item to load
		@return The <code>Message</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Message getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Message m = dc.loadItem(Message.class, id);
		if (m == null) throw new ItemNotFoundException("Message[id="+id+"]");
		return m;
	}

	/**
		Count the number of unread messages for the specified or logged in user.
		The count doesn't include messages which have been flagged as removed.
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param user The user to count the unread messages for, or null to
			count the messages for the logged in user
		@param cacheResult If the result should be cached or not
		@return The number of unread messages
		@throws BaseException If there is an error
	*/
	public static long countUnreadMessages(DbControl dc, User user, boolean cacheResult)
		throws BaseException
	{
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"COUNT_UNREAD_MESSAGES_FOR_USER");
		/*
			SELECT count(*) 
			FROM MessageData msg
			WHERE msg.to = :user AND msg.read = false AND msg.removed = false
		*/
		query.setInteger("user", user == null ? dc.getSessionControl().getLoggedInUserId() : user.getId());
		query.setCacheable(cacheResult);
		return HibernateUtil.loadData(Long.class, query);
	}
	
	/**
		Get a query configured to retrieve messages for the specified user.  
		@param user The user to retreive messages for, null is allowed if
			the logged in user has generic read permission for messages in which case
			all messages will be returned
		@return An {@link ItemQuery} object
		@see User#getMessages()
	*/
	public static ItemQuery<Message> getQuery(User user)
	{
		ItemQuery<Message> query = null;
		
		if (user != null)
		{
			query = new ItemQuery<Message>(Message.class, null);
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("to"), 
					Hql.entity(user)
				)
			);
		}
		else
		{
			query = new ItemQuery<Message>(Message.class);
		}
		return query;
	}

	private boolean systemCreated;
	private boolean sendAsEmail;
	
	Message(MessageData messageData)
	{
		super(messageData);
	}


	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	public String getName()
	{
		return getData().getName();
	}
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	public String getDescription()
	{
		return getData().getDescription();
	}
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	public boolean isRemoved()
	{
		return getData().isRemoved();
	}
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemoved(removed);
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Read, write an delete permission is granted if the logged in user
		is the same as the user the message is to.
	*/
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (getSessionControl().getLoggedInUserId() == getData().getTo().getId())
		{
			granted |= Permission.grant(Permission.READ, Permission.WRITE, Permission.DELETE);
		}
		if (systemCreated) granted |= Permission.grant(Permission.CREATE);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Send the message as an email if requested by the user.
	*/
	@Override
	void onAfterCommit(Transactional.Action action)
	{
		super.onAfterCommit(action);
		
		if (action == Transactional.Action.CREATE)
		{
			if (sendAsEmail) 
			{
				try
				{
					this.sendAsEmail();
				}
				catch (RuntimeException ex)
				{
					Application.getLogger().warn(ex.getMessage(), ex);
				}
			}
		}
		
	}


	/**
		Get the {@link User} this message was sent to.
		@return The <code>User</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the user
		@throws BaseException If there is another error
	*/
	public User getTo()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(User.class, getData().getTo());
	}
	
	/**
		Set the user this message should be sent to.
		@throws InvalidDataException If the user is null
	*/
	private void setTo(User to)
		throws InvalidDataException
	{
		if (to == null) throw new InvalidUseOfNullException("to");
		getData().setTo(to.getData());
		sendAsEmail = EmailUtil.isEnabled(to);
	}
	
	/**
		Get the name of the sender of this message. The sender may be another user or
		a system process.
		@see #getSender()
	*/
	public String getFrom()
	{
		return getData().getFrom();
	}
	
	/**
		The maximum length of the sender name.
		@see #getNew(DbControl, User, String, User, Job)
	*/
	public static final int MAX_FROM_LENGTH = MessageData.MAX_FROM_LENGTH;
	/**
		Set the name of the sender of the message.
		@throws InvalidDataException If the sender is null or longer than
			{@link #MAX_FROM_LENGTH}
	*/
	private void setFrom(String from)
		throws InvalidDataException
	{
		getData().setFrom(StringUtil.setNotNullString(from, "from", MAX_FROM_LENGTH));
	}
	
	/**
		Get the user that sent this message.
		@return A <code>User</code> object or null if the message was sent by the system or
			the sending user account has been deleted
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the sender
		@throws BaseException If there is another error
		@see #getFrom()
	*/
	public User getSender()
		throws PermissionDeniedException, BaseException
	{
		int fromId = getData().getFromId();
		User sender = null;
		if (fromId != 0)
		{
			try
			{
				sender = User.getById(getDbControl(), fromId);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		return sender;
	}
	
	private void setSender(User sender)
	{
		getData().setFromId(sender == null ? 0 : sender.getId());
	}
	
	/**
		Get the {@link Job} this message is about.
		@return The <code>Job</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the job
		@throws BaseException If there is another error
	*/
	public Job getJob()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Job.class, getData().getJob());
	}
	private void setJob(Job job)
		throws PermissionDeniedException
	{
		if (job != null) job.checkPermission(Permission.READ);
		getData().setJob(job == null ? null : job.getData());
	}
	
	/**
		Get the date and time the message was sent.
	*/
	public Date getTimeSent()
	{
		return DateUtil.copy(getData().getTimeSent());
	}
	
	/**
		Check if the message has been read or not.
	*/
	public boolean isRead()
	{
		return getData().isRead();
	}
	
	/**
		Set read status of the message.
		@param isRead TRUE to mark the message as read, FALSE to mark it as unread
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
	*/
	public void setRead(boolean isRead)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setRead(isRead);
	}
	
	/**
		Send this message as an email. This method call is ignored
		if the server admin has not configured an outgoing SMTP
		server. 
		@since 2.16
		@see EmailUtil#isEnabled()
	*/
	public void sendAsEmail()
	{
		if (!EmailUtil.isEnabled()) return;
		try
		{
			Email email = EmailUtil.createSimpleEmail();
			UserData user = getData().getTo();
			email.addTo(user.getEmail(), user.getName(), "UTF-8");
			email.setSubject(getName());
			String message = getDescription();
			if (message != null && message.length() > 0)
			{
				email.setMsg(message);
			}
			email.send();
		}
		catch (EmailException ex)
		{
			throw new BaseException("Could not send email for message:" + this, ex);
		}
	}
	
}
