/**
	$Id: SchemaGenerator.java 5427 2010-09-24 09:24:49Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;

import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.dbengine.DbEngine;

import org.hibernate.cfg.Configuration;
import org.hibernate.dialect.Dialect;
import org.hibernate.jdbc.Work;
import org.hibernate.tool.hbm2ddl.DatabaseMetadata;

/**
	Replaces the Hibernate SchemaExport and SchemaUpdate tools. We need
	this to be able to intercept and modify some of the SQL statments that
	Hibernate generates since they don't always work as expected.
	In particular, since Hibernate 3.5 they generate "not null" constraints for
	columns that are added to existing tables and we don't want that.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2010-09-24 11:24:49 +0200 (Fri, 24 Sep 2010) $
*/
public class SchemaGenerator
	implements Work
{
	
	private static final org.apache.log4j.Logger log = 
		org.apache.log4j.LogManager.getLogger("net.sf.basedb.core.hibernate.SchemaGenerator");
	
	private final Configuration cfg;
	private final Dialect dialect;
	private final DbEngine dbEngine;
	private final boolean update;
	private final ProgressReporter progress;

	/**
		Create a new schema generator. 
		
		@param cfg The current configuration
		@param dialect The dialect of the database we are connecting to
		@param dbEngine The DbEngine for the database we are connecting to
		@param update TRUE if we are updating an existing database, FALSE if
			we are creating a new one
		@param progress An optional progress reporter
	*/
	public SchemaGenerator(Configuration cfg, Dialect dialect, DbEngine dbEngine, boolean update, ProgressReporter progress)
	{
		this.cfg = cfg;
		this.dialect = dialect;
		this.dbEngine = dbEngine;
		this.update = update;
		this.progress = progress;
	}

	/*
		From the Work interface
		-----------------------
	*/
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		log.info("Starting schema generator");
		log.info("Dialect: " + dialect);
		log.info("DbEngine: " + dbEngine);
		
		Statement stmt = connection.createStatement();
		try
		{
			String[] allSql = null;
			if (update)
			{
				log.info("Fetching database metadata");
				DatabaseMetadata meta = new DatabaseMetadata(connection, dialect);
				allSql = cfg.generateSchemaUpdateScript(dialect, meta);
			}
			else
			{
				allSql = cfg.generateSchemaCreationScript(dialect);
			}
			int numStatements = allSql.length;
			int numDone = 0;
			int interval = Math.max(1, numStatements / 30);
			for (String sql : allSql)
			{
				String actualSql = dbEngine.inspectSchemaGenerationSQL(sql, dialect, update);
				if (actualSql != null)
				{
					if (!actualSql.equals(sql))
					{
						log.debug("Original SQL: " + sql);
					}
					log.debug("Executing: " + actualSql);
					stmt.executeUpdate(actualSql);
				}
				else
				{
					log.debug("Ignoring: " + sql);
				}
				if (progress != null)
				{
					if ((numDone % interval == 0) && (actualSql != null))
					{
						progress.display((100 * numDone) / numStatements, "--" + StringUtil.trimString(actualSql, 45));
					}
					else
					{
						progress.append(".");
					}
				}
				numDone++;
			}
		}
		catch (SQLException ex)
		{
			log.error("Schema update failed", ex);
			throw ex;
		}
		finally
		{
			if (stmt != null) 
			{
				try
				{
					stmt.close();
				}
				catch (SQLException ex)
				{}
			}
		}
		log.info("Schema update complete" );
		
	}

}
