/*
	$Id: AutoDetectFileFormat.java 5588 2011-03-15 13:53:10Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.SessionControl;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Location;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.PluginConfiguration;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.plugin.AutoDetectingImporter;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.util.ContextUtil.ContextResult;
import net.sf.basedb.util.filter.Filter;

import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.LinkedHashMap;
import java.util.Collections;
import java.io.InputStream;

/**
	This class has methods for autodetecting file formats.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2011-03-15 14:53:10 +0100 (Tue, 15 Mar 2011) $
*/
public class AutoDetectFileFormat
{

	/**
		Same as {@link #findPlugins(DbControl, GuiContext, File, String, PluginDefinition)}
		with no value for the charset parameter.
		@deprecated
	*/
	public static Map<PluginDefinition, List<PluginConfiguration>> findPlugins(DbControl dc, 
		GuiContext context, File file, PluginDefinition pluginDef)
		throws InvalidDataException, BaseException
	{
		return findPlugins(dc, context, file, null, pluginDef);
	}

	/**
		Find all plugins with configurations which can import the specified file. 
		This method will check all plugins implementing the {@link
		AutoDetectingImporter} interface if they can import the specified file.
		
		@param dc The <code>DbControl</code> object to use for database access
		@param context A required context for the plugins (ie. what items to import)
		@param file The file to import
		@param charset If given, overrides the character set specified by the file
			{@link File#getCharacterSet()}
		@param pluginDef Check only the configurations of this plugin definition, or
			null if all plugin definitions should be checked
		@return A map which for each plugin definition contains a list
			of mathching configurations
		@throws InvalidDataException If the file parameter is null or if
			the file isn't located in {@link Location#PRIMARY} storage
		@throws BaseException If there is another error
		@since 2.9
	*/
	public static Map<PluginDefinition, List<PluginConfiguration>> findPlugins(DbControl dc, 
		GuiContext context, File file, String charset, PluginDefinition pluginDef)
		throws InvalidDataException, BaseException
	{
		if (file == null) throw new InvalidUseOfNullException("file");
		if (!file.getLocation().isDownloadable())
		{
			throw new InvalidDataException("Can't download file data for file '" + 
					file.getName() + "'; location=" + file.getLocation());
		}
		SessionControl sc = dc.getSessionControl();
		Object currentItem = null;
		if (context.getType() == GuiContext.Type.ITEM)
		{
			currentItem = context.getItem().getById(dc, sc.getCurrentContext(context.getItem()).getId());
		}

		Map<PluginDefinition, List<PluginConfiguration>> plugins = 
			new LinkedHashMap<PluginDefinition, List<PluginConfiguration>>();
		
		List<PluginDefinition> pluginDefs = null;
		if (pluginDef == null)
		{
			ItemQuery<PluginDefinition> pluginQuery = 
				PluginDefinition.getQuery(context, AutoDetectingImporter.class.getName());
			pluginQuery.order(Orders.asc(Hql.property("name")));
			pluginQuery.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			pluginDefs = pluginQuery.list(dc);
		}
		else
		{
			pluginDefs = Collections.singletonList(pluginDef);
		}
		
		for (PluginDefinition plugin : pluginDefs)
		{
			List<PluginConfiguration> configs = new LinkedList<PluginConfiguration>();
			if (!plugin.requiresConfiguration())
			{
				if (checkImportable(sc, plugin, null, file, charset, context, currentItem))
				{
					configs.add(null);
				}
			}
			if (plugin.supportsConfigurations())
			{
				ItemQuery<PluginConfiguration> configQuery = plugin.getPluginConfigurations();
				configQuery.restrict(Restrictions.gt(Hql.property("parameterVersion"), Expressions.integer(0)));
				configQuery.order(Orders.asc(Hql.property("name")));
				configQuery.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
				for (PluginConfiguration config : configQuery.list(dc))
				{
					if (checkImportable(sc, plugin, config, file, charset, context, currentItem))
					{
						configs.add(config);
					}
				}
			}
			if (configs.size() > 0)	plugins.put(plugin, configs);
		}
		return plugins;
	}
	
	private static boolean checkImportable(SessionControl sc, PluginDefinition plugin, 
		PluginConfiguration config, File file, String charset, GuiContext context, Object currentItem)
	{
		AutoDetectingImporter importer = null;
		boolean canBeUsed = false;
		try
		{
			importer = plugin.newInstance(AutoDetectingImporter.class, null, sc, config, null);
			boolean isInContext = true;
			if (plugin.isInteractive() && context != null)
			{
				try
				{
					InteractivePlugin ip = (InteractivePlugin)importer;
					isInContext = ip.isInContext(context, currentItem) == null;
				}
				catch (Throwable t)
				{
					isInContext = false;
				}
			}
			if (isInContext)
			{
				InputStream in = new FileImportInputStream(file, 0, charset);
				try
				{
					canBeUsed = importer.isImportable(in);
				}
				catch (Throwable t)
				{
					canBeUsed = false;
				}
				finally
				{
					try
					{
						in.close();
					}
					catch (Throwable t)
					{}
				}
			}			
		}
		finally
		{
			if (importer != null) importer.done();
		}
		return canBeUsed;
	}
	
	/**
		Filter implementation that filters a collection of
		{@link ContextResult}:s by checking each plug-ins ability to
		import a given file. Eg. the ContextResult is accepted if the
		{@link AutoDetectingImporter#isImportable(InputStream)} returns
		true. It is assumed that the collection of ContextResult:s only
		contains import plug-ins that implements the AutoDetectingImporter 
		interface.
	
		@author Nicklas
		@version 2.9
		@base.modified $Date: 2011-03-15 14:53:10 +0100 (Tue, 15 Mar 2011) $
	*/
	public static class IsImportableFilter
		implements Filter<ContextResult>
	{
		private File file;
		private Filter<? super ContextResult> parent;
		
		public IsImportableFilter(File file, Filter<? super ContextResult> parent)
		{
			this.file = file;
			this.parent = parent;
		}
		
		public boolean evaluate(ContextResult result)
		{
			if (result == null) return false;
			if (parent != null && !parent.evaluate(result)) return false;
			if (!result.isInContext()) return false;
			return checkImportable(file.getSessionControl(), 
					result.getPluginDefinition(), result.getPluginConfiguration(), file, 
					null, null, null);
		}
	}
	
}
