/*
	$Id: FileUtil.java 5384 2010-08-13 08:50:27Z nicklas $

	Copyright (C) 2005, 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Directory;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.SimpleAbsoluteProgressReporter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.filter.Filter;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PushbackInputStream;
import java.io.Reader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
	This class collects some useful methods for file and stream
	handling.

	@author Nicklas
	@version 2.0
	@base.modified $Date$
*/
public class FileUtil
{
	private static final int BUFFER_SIZE = 512;

	/**
		Copy from the input stream to the output stream
		until end of file is reached.
		@param in The <code>InputStream</code> to read from
		@param out The <code>OutputStream</code> to write to
		@return The number of bytes copied
		@throws IOException This exception is thrown if there is an error
	*/
	public static long copy(InputStream in, OutputStream out)
		throws IOException
	{
		return copy(in, out, null);
	}

	/**
		Copy from the input stream to the output stream
		until end of file is reached.
		@param in The <code>InputStream</code> to read from
		@param out The <code>OutputStream</code> to write to
		@param progress An optional progress reporter
		@return The number of bytes copied
		@throws IOException This exception is thrown if there is an error
		@since 2.5
	*/
	public static long copy(InputStream in, OutputStream out, AbsoluteProgressReporter progress)
		throws IOException
	{
		int bytes = 0;
		long totalBytes = 0;
		byte[] buffer = new byte[BUFFER_SIZE];
	
		while (bytes != -1) // -1 = end of stream
		{
			ThreadSignalHandler.checkInterrupted();
			bytes = in.read(buffer, 0, buffer.length);
			if (bytes > 0)
			{
				totalBytes += bytes;
				out.write(buffer, 0, bytes);
				if (progress != null) progress.displayAbsolute(totalBytes, null);
			}
		}
		out.flush();
		return totalBytes;
	}

	/**
		Copy from the reader to the writer until end of file is reached.
		@param in The <code>Reader</code> to read from
		@param out The <code>Writer</code> to write to
		@return The number of characters copied
		@throws IOException This exception is thrown if there is an error
		@since 2.12
	*/
	public static long copy(Reader in, Writer out)
		throws IOException
	{
		return copy(in, out, null);
	}
	
	/**
		Copy from the reader to the writer until end of file is reached.
		@param in The <code>Reader</code> to read from
		@param out The <code>Writer</code> to write to
		@param progress An optional progress reporter
		@return The number of characters copied
		@throws IOException This exception is thrown if there is an error
		@since 2.12
	*/
	public static long copy(Reader in, Writer out, AbsoluteProgressReporter progress)
		throws IOException
	{
		int numRead = 0;
		long totalRead = 0;
		char[] buffer = new char[BUFFER_SIZE];
	
		while (numRead != -1) // -1 = end of stream
		{
			ThreadSignalHandler.checkInterrupted();
			numRead = in.read(buffer, 0, buffer.length);
			if (numRead > 0)
			{
				totalRead += numRead;
				out.write(buffer, 0, numRead);
				if (progress != null) progress.displayAbsolute(totalRead, null);
			}
		}
		return totalRead;
	}
	
	/**
		Read from the input stream until the end is reached.
		
		@param in The <code>InputStream</code> to read from
		@return The number of bytes copied
		@throws IOException This exception is thrown if there is an error
		@since 2.4
	*/
	public static long read(InputStream in)
		throws IOException
	{
		int bytes = 0;
		long totalBytes = 0;
		byte[] buffer = new byte[BUFFER_SIZE];
	
		while (bytes != -1) // -1 = end of stream
		{
			ThreadSignalHandler.checkInterrupted();
			bytes = in.read(buffer, 0, buffer.length);
			if (bytes > 0)
			{
				totalBytes += bytes;
			}
		}
		return totalBytes;
	}

	/**
		Close an input stream without throwing an exception.
		@param in The input stream to close
		@since 2.9
	*/
	public static void close(InputStream in)
	{
		if (in == null) return;
		try
		{
			in.close();
		}
		catch (Throwable t)
		{}
	}
	
	/**
		Close an output stream without throwing an exception.
		@param out The output stream to close
		@since 2.9
	*/
	public static void close(OutputStream out)
	{
		if (out == null) return;
		try
		{
			out.close();
		}
		catch (Throwable t)
		{}
	}
	
	/**
		Get a buffered <code>InputStream</code> object reading from
		the specified file.

		@param file The path to the file to read from
		@return A buffered <code>InputStream</code>
		@throws IOException If the stream cannot be opened
	*/
	public static InputStream getInputStream(File file)
		throws IOException
	{
		return new BufferedInputStream(new FileInputStream(file));
	}
	
	/**
		Get a buffered <code>OutputStream</code> object writing to
		the specified file.
	
		@param file The path to the file to write to
		@return A buffered <code>OutputStream</code>
		@throws IOException If the stream cannot be opened
	*/
	public static OutputStream getOutputStream(File file)
		throws IOException
	{
		return new BufferedOutputStream(new FileOutputStream(file));
	}
	
	/**
		Take a peek at an input stream and check if the first few bytes
		matches the bCheck parameter. After the check the bytes are
		pushed back to the stream again. This method is useful to check
		the file format of a file, which can often be done by checing
		the first few bytes at the start of the file. For example
		to check if the file is gzip file:
		<p>
		<code>bCheck = new byte[] { 0x1f, (byte)0x8b };</code>
		
		@param pin The input stream to read from which must have
			large enough buffer to be able to unread the bytes
		@param bCheck The byte values to use for comparison
		@return TRUE if the file matches the specified bytes, 
			FALSE otherwise
		@throws IOException If there is an error reading or unreading
			the bytes
		@throws NullPointerException If pin or bCheck is null
		@throws IllegalArgumentException If bCheck is zero length
		@since 2.4
	*/
	public static boolean checkMagicNumber(PushbackInputStream pin, byte[] bCheck)
		throws IOException
	{
		if (pin == null) throw new NullPointerException("pin");
		if (bCheck == null) throw new NullPointerException("bCheck");
		if (bCheck.length == 0) throw new IllegalArgumentException("bCheck has zero length");
		byte[] b = new byte[bCheck.length];
		int numRead = pin.read(b);
		boolean match = numRead == bCheck.length && Arrays.equals(b, bCheck);
		if (numRead > 0) pin.unread(b, 0, numRead);
		return match;
	}
	
	/**
		Get the complete tree of sub-directories from a given directory.
		<p>
		<b>NOTE!!! This method has very bad performance on large directory trees.
		Consider if it is not possible to use the {@link 
		#loadMinimalDirectoryTree(DbControl, Directory...)}
		instead together with lazy loading of subdirectories.</b>
		
	 	@param dc DbControl used to access the database. 
		@param directory The directory to start with
		@return A <code>Map</code> which maps a directory to a 
			list of it's sub-directories. Directories without sub-directories
			are not included in the map.
		@throws InvalidDataException If the directory is null
		@throws BaseException If there is another error
	*/
	public static Map<Directory, List<Directory>> getDirectoryTree(DbControl dc, Directory directory)
		throws InvalidDataException, BaseException
	{
		ItemQuery<Directory> query = Directory.getQuery();
		query.restrict(
			Restrictions.in(
				Hql.property("parent"),
				Expressions.parameter("parents")
			)
		);
		query.order(Orders.asc(Hql.property("name")));
		query.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.NOT_REMOVED, Include.REMOVED);
		
		Map<Directory, List<Directory>> tree = new HashMap<Directory, List<Directory>>();
		List<Integer> parents = new LinkedList<Integer>();
		parents.add(directory.getId());
		while (parents.size() > 0)
		{
			query.setParameter("parents", parents, Type.INT);
			List<Directory> children = query.list(dc);
			parents.clear();
			for (Directory child : children)
			{
				Directory parent = child.getParent();
				if (!tree.containsKey(parent))
				{
					tree.put(parent, new LinkedList<Directory>());
				}
				tree.get(parent).add(child);
				parents.add(child.getId());
			}
		}
		return tree;
	}
	
	/**
		Load a minimal directory tree that includes at least all the specified directories 
		and their parent directories all the way up to the root or as far as the logged in user
		has read permission.
		@param dc A DbControl for accessing the database
		@param directories An array of directories
		@return A <code>Map</code> which maps a directory to a 
			list of it's sub-directories. The map is guaranteed to contain an entry for each of
			the given directories, but the list may be empty if there are no subdirectories
		@throws BaseException If there is an error
		@since 2.11
	*/
	public static Map<Directory, List<Directory>> loadMinimalDirectoryTree(DbControl dc, Directory... directories)
		throws InvalidDataException, BaseException
	{
		Map<Directory, List<Directory>> tree = new HashMap<Directory, List<Directory>>();

		// 1. Move up the directory tree as far as possible
		//    storing all directories that we pass
		Set<Integer> all = new HashSet<Integer>();
		for (Directory d : directories)
		{
			if (d == null)
			{
				/* #### CONTINUE-STATEMENT #### */
				continue;
			}
			tree.put(d, new LinkedList<Directory>());
			all.add(d.getId());
			try
			{
				Directory parent = d.getParent();
				while (parent != null)
				{
					if (all.add(parent.getId()))
					{
						parent = parent.getParent();
					}
					else
					{
						// break if the directory has already been added
						parent = null;
					}
				}
			}
			catch (PermissionDeniedException ex)
			{}
		}

		// 2. Query that loads the subdirectories to all directories that we passed
		ItemQuery<Directory> query = Directory.getQuery();
		query.restrict(
			Restrictions.in(
				Hql.property("parent"),
				Expressions.parameter("parents")
			)
		);
		query.order(Orders.asc(Hql.property("name")));
		query.include(Include.MINE, Include.SHARED, Include.OTHERS, Include.NOT_REMOVED, Include.REMOVED);
		query.setParameter("parents", all, Type.INT);
		
		ItemResultIterator<Directory> it = query.iterate(dc);
		while (it.hasNext())
		{
			Directory child = it.next();
			Directory parent = child.getParent();
			if (!tree.containsKey(parent))
			{
				tree.put(parent, new LinkedList<Directory>());
			}
			tree.get(parent).add(child);
		}
		return tree;
	}
	
	
	/**
	 	Find files matching a given filter in a directory. If the filter
	 	returns sub-directories, this method will check those directories too.
	 	Use, for example, the {@link RegexpFileFilter} to look for files
	 	matching a regular expression.
	 	
	  	@param directory The parent directory to search in
	  	@param filter A filter which specifies which files to find, or null
	  		to find all files
	  	@return A List with the files in directory
	  		that matches the filter, or null if the specified directory doesn't
	  		exists or isn't a directory
	  	@since 2.5
	  	@see RegexpFileFilter
	*/
	public static List<File> findFiles(File directory, FileFilter filter)
	{	
		if (!directory.exists() || !directory.isDirectory())
		{
			return null;
		}
		
		// To hold the result
		List<File> searchResult = new LinkedList<File>();
		
		// Temporary holds the directories too look in
		List<File> directories = new LinkedList<File>();
		directories.add(directory);
		
		while (directories.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			File dir = directories.remove(0);
			File[] files = dir.listFiles(filter);
			for (File f : files)
			{
				ThreadSignalHandler.checkInterrupted();
				if (f.isDirectory())
				{
					directories.add(f);
				}
				else
				{
					searchResult.add(f);
				}
			}
		}
		return searchResult;
	}
	
	/**
		Deletes all empty subdirectories and, optionally, also
		the root directory.
		@param root The root directory to start looking in
		@param deleteRoot TRUE to also delete the root directory
			if it is empty, FALSE to leave it
		@return The number of deleted directories
		@since 2.13
	*/
	public static int deleteEmptyDirectories(File root, boolean deleteRoot)
	{
		// Step 1 - Create a list with all subdirectories
		List<File> allDirs = new LinkedList<File>();
		List<File> unchecked = new LinkedList<File>();
		unchecked.add(root);
		while (unchecked.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			File dir = unchecked.remove(0);
			File[] files = dir.listFiles();
			if (files != null)
			{
				for (File f : files)
				{
					if (f.isDirectory()) unchecked.add(f);
				}
				// Important! Add the directory first in the list!
				if (deleteRoot || !dir.equals(root)) allDirs.add(0, dir);
			}
		}
		
		// Step 2 - Remove empty directories
		int numDeleted = 0;
		for (File subDir : allDirs)
		{
			ThreadSignalHandler.checkInterrupted();
			File[] files = subDir.listFiles();
			if (files.length == 0) 
			{
				if (subDir.delete()) numDeleted++;
			}
		}
		return numDeleted;
	}
	
	/**
		Do something recursively to all files and sub-directories in
		specified root directory. If no directory filter is specified,
		all subdirectories are processed, otherwise only those directories
		were the {@link Filter#evaluate(Object)} method returns TRUE.
		<p>
		Files are only loaded if a file filter is specified. The
		return value of the <code>evaluate</code> method is ignored.
		<p>
		The files and directories may be processed in any order.
		<p>
		NOTE! The intention of this method is that the filter should
		have have a "side-effect" that does some work on the files/
		sub-directories.
		
		@param dc A DbControl to use for database access
		@param root The root directory
		@param dirFilter A filter that selects/rejects sub-directories
		@param fileFilter	A filter that does something with the file
		@since 2.13
	*/
	public static void doRecursively(DbControl dc, Directory root, 
			Filter<? super Directory> dirFilter, Filter<? super net.sf.basedb.core.File> fileFilter)
	{
		// Create the file/directory queries
		ItemQuery<Directory> dirQuery = Directory.getQuery();
		dirQuery.restrict(
				Restrictions.eq(
					Hql.property("parent.id"), 
					Expressions.parameter("parent")
				)
			);
		dirQuery.include(Include.ALL);
		
		ItemQuery<net.sf.basedb.core.File> fileQuery = net.sf.basedb.core.File.getQuery();
		fileQuery.restrict(
				Restrictions.eq(
					Hql.property("directory.id"), 
					Expressions.parameter("directory")
				)
			);
		fileQuery.include(Include.ALL);
		
		// Load the staring directories
		List<Directory> directories = new LinkedList<Directory>();
		directories.add(root);
		
		while (directories.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			Directory d = directories.remove(0);
			if (dirFilter == null || dirFilter.evaluate(d)) 
			{
				dirQuery.setParameter("parent", d.getId(), Type.INT);
				directories.addAll(dirQuery.list(dc));
			}
			if (fileFilter != null)
			{
				fileQuery.setParameter("directory", d.getId(), Type.INT);
				for (net.sf.basedb.core.File f : fileQuery.list(dc))
				{
					fileFilter.evaluate(f);
				}
			}
		}
	}
	
	/**
		Creates a temporary directory. This method uses the built-in {@link
		File#createTempFile(String, String, File)} method to first crete a temporary
		file, that is then deleted and made into a directory instead. This should
		ensure that a new empty directory is created each time this method is called.
		The "temporary" nature of this method doesn't involve cleaning up or removing
		the created directory. It is the responsibility of the calling code to remove
		every file and subdirectory as well as the returned directory. This can
		for example be done with the {@link #deleteTempDirectory(File)} method.
		
		@param prefix A prefix that is used in the directory name 
			it must be at least 3 characters long.
		@param suffix An optional suffix for the directory name
		@param directory An optional parent directory in which the temporary
			directory should be created, or null to use the system temporary 
			directory
		@return A reference to the created directory
		@throws IOException If there is a problem with the directory creation
		@since 2.15
		@see File#createTempFile(String, String, File)
	*/
	public static File createTempDirectory(String prefix, String suffix, File directory)
		throws IOException
	{
		File tmpDir = File.createTempFile(prefix, suffix, directory);
		if (!tmpDir.delete() || !tmpDir.mkdir())
		{
			throw new IOException("Could not create directory: " + tmpDir.getAbsolutePath());
		}
		return tmpDir;
	}
	
	/**
		Recursively delete all files and subdirectories within a given
		directory and the also the given directory itself.

		@param root The root directory to start looking in
		@return The number of deleted files and directories
		@since 2.15
	*/
	public static int deleteTempDirectory(File root)
	{
		int numDeleted = 0;

		// Step 1 - Remove the files and create a list with all subdirectories
		List<File> allDirs = new LinkedList<File>();
		List<File> unchecked = new LinkedList<File>();
		unchecked.add(root);
		while (unchecked.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			File dir = unchecked.remove(0);
			File[] files = dir.listFiles();
			if (files != null)
			{
				// Important! Add the directory first in the list!
				allDirs.add(0, dir);
				for (File f : files)
				{
					if (f.isDirectory()) 
					{
						unchecked.add(f);
					}
					else
					{
						if (f.delete()) numDeleted++;
					}
				}
			}
		}
		
		// Step 2 - Remove the directories
		for (File subDir : allDirs)
		{
			ThreadSignalHandler.checkInterrupted();
			if (subDir.delete()) numDeleted++;
		}
		return numDeleted;
	}
	
	/**
		Upload multiple files to BASE. 
		@param toDir The directory in the BASE file system that the files should
			be uploaded to
		@param fromDir The directory in the local file system to scan for files 
			to upload if the directory doesn't exists or is empty an empty list is
			returned
		@param filter An optional filter that decides which files that should
			be uploaded or not, if null all files, including subdirectories, are
			uploaded
		@param overwrite TRUE if existing files should be overwritten, FALSE to
			only overwrite files that have been marked for removal
		@param progress An optional progress reporter
		@return A list with the new file items that were created
		@since 2.16
	*/
	public static List<net.sf.basedb.core.File> uploadFiles(DbControl dc, Directory toDir, 
		File fromDir, final FileFilter filter, boolean overwrite, ProgressReporter progress)
		throws IOException
	{
		if (!fromDir.exists() || !fromDir.isDirectory())
		{
			return Collections.emptyList();
		}
	
		// All files to upload in BASE and native representation
		List<net.sf.basedb.core.File> uploaded = new ArrayList<net.sf.basedb.core.File>();
		List<File> fileToUpload = new ArrayList<File>();

		// Temporary holds the directories too look in
		List<File> directories = new LinkedList<File>();
		Map<File, Directory> subdirs = new HashMap<File, Directory>();
		directories.add(fromDir);
		subdirs.put(fromDir, toDir);
		
		// First step: Scan the directories to find all files to upload
		// This step will create BASE File and Directory items
		long totalBytes = 0;
		while (directories.size() > 0)
		{
			ThreadSignalHandler.checkInterrupted();
			File dir = directories.remove(0);
			Directory uploadDir = subdirs.get(dir);
			
			File[] files = dir.listFiles(filter);
			for (File f : files)
			{
				if (f.isDirectory())
				{
					// Create subdirectory
					directories.add(f);
					Directory newDirectory = Directory.getNew(dc, uploadDir);
					newDirectory.setName(f.getName());
					dc.saveItem(newDirectory);
					subdirs.put(f, newDirectory);
				}
				else if (f.isFile())
				{
					// Upload the file
					net.sf.basedb.core.File baseFile = net.sf.basedb.core.File.getFile(dc, uploadDir, f.getName(), true);
					if (!baseFile.isInDatabase())
					{
						dc.saveItem(baseFile);
					}
					if (overwrite || !baseFile.isInDatabase() || baseFile.isRemoved())
					{
						uploaded.add(baseFile);
						fileToUpload.add(f);
						totalBytes += f.length();
					}
				}
			}
		}

		// Second step: Perform the upload
		SimpleAbsoluteProgressReporter aProgress = null;
		if (progress != null) aProgress = new SimpleAbsoluteProgressReporter(progress, totalBytes);
		long numBytes = 0;
		for (int i = 0; i < fileToUpload.size(); ++i)
		{
			ThreadSignalHandler.checkInterrupted();
			File f = fileToUpload.get(i);
			net.sf.basedb.core.File baseFile = uploaded.get(i);
			InputStream in = null;
			OutputStream out = null;
			try
			{
				if (aProgress != null) 
				{
					aProgress.setAbsolute(numBytes, "Uploading: " + f.getName());
				}
				baseFile.setRemoved(false);
				baseFile.setMimeTypeAuto(null, null);
				in = FileUtil.getInputStream(f);
				out = baseFile.getUploadStream(false);
				numBytes += FileUtil.copy(in, out, aProgress);
			}
			finally
			{
				FileUtil.close(in);
				FileUtil.close(out);
			}
		}
		return uploaded;
	}
	
}
