/**
	$Id: LabeledExtractToExtractTransformer.java 4723 2009-01-13 08:20:26Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.biomaterial;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.collections.BasicItemToIdTransformer;
import net.sf.basedb.util.collections.CollectionTransformer;
import net.sf.basedb.util.collections.MultiStepCollectionTransformer;

/**
	Collection transformer that given a collection of labeled extracts finds
	their parent extract. Note that this transformer is only capable 
	of finding the extracts that are immediate
	parents to the labeled extract. If you want to follow pooled labeled 
	extracts up to extracts, you should combine this transformer
	and a {@link PooledParentsTransformer} with the help of a
	{@link MultiStepCollectionTransformer}.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2009-01-13 09:20:26 +0100 (Tue, 13 Jan 2009) $
*/
public class LabeledExtractToExtractTransformer
	implements CollectionTransformer<LabeledExtract, Extract>
{

	private DbControl dc;
	public LabeledExtractToExtractTransformer(DbControl dc)
	{
		this.dc = dc;
	}
	
	/*
		From the CollectionTransformer interface
		----------------------------------------
	*/
	/**
		Add all samples that are parents to any of the extracts in the
		source collection.
		@return The number of samples added
	*/
	@Override
	public int transform(Collection<? extends LabeledExtract> labeledExtracts,
			Collection<? super Extract> extracts)
	{
		// Copy source ID:s so we can use them as a parameter in the query
		Set<Integer> sourceIds = new HashSet<Integer>();
		BasicItemToIdTransformer idTransformer = new BasicItemToIdTransformer();
		idTransformer.transform(labeledExtracts, sourceIds);

		ItemQuery<Extract> query = createQuery();
		query.join(Hql.innerJoin("children", Item.LABELEDEXTRACT.getAlias()));
		query.restrict(
			Restrictions.in(
				Hql.alias(Item.LABELEDEXTRACT.getAlias()), 
				Expressions.parameter("labledExtracts")
			)
		);
		query.setParameter("labledExtracts", sourceIds, Type.INT);
		int size = extracts.size();
		extracts.addAll(query.list(dc));
		return extracts.size() - size;
	}
	// -------------------------------------------------
	
	/**
		Create a query for loading extracts. The default implemenation will
		simply load all extracts. The restriction needed to locate the child
		labeled extracts is automatically applied by the <code>transform</code>
		method. Subclasses may override this method to create an alternate query
		(for example, with more filters).
		@return A query object
	*/
	protected ItemQuery<Extract> createQuery()
	{
		ItemQuery<Extract> query = Extract.getQuery();
		query.include(Include.ALL);
		return query;
	}
	
}
