/**
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.export.spotdata;

import java.util.ArrayList;
import java.util.List;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ExtendedProperties;
import net.sf.basedb.core.ExtendedProperty;
import net.sf.basedb.core.ExtraValue;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.Formula;
import net.sf.basedb.core.Include;
import net.sf.basedb.core.IntensityTransform;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.RawDataProperty;
import net.sf.basedb.core.RawDataType;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.ExtraValue.CoordinateType;
import net.sf.basedb.core.Metadata.PropertyPath;
import net.sf.basedb.core.data.RawData;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.snapshot.SnapshotManager;

/**
	Abstract helper class for implementing {@link ExportableFieldConverter}.
	It contains a lot of helper method for locating formulas, extra values,
	experimental factors, etc.
	
	@author nicklas
	@since 2.15
*/
public abstract class AbstractFieldConverter
	implements ExportableFieldConverter
{
	private final DbControl dc;
	private final BioAssaySet source;

	private SnapshotManager snapshotManager;
	private ItemQuery<AnnotationType> factorQuery;
	private ItemQuery<AnnotationType> assayAnnotationQuery;
	private ItemQuery<ExtraValue> extraQuery;
	private ItemQuery<Formula> formulaQuery;

	/**
		Creates a new field converter to be used with a specific 
		bioassay set source.
		@param dc A DbControl to use for database access
		@param source The source bioassay set
	*/
	protected AbstractFieldConverter(DbControl dc, BioAssaySet source)
	{
		if (dc == null) throw new NullPointerException("dc");
		if (source == null) throw new NullPointerException("source");
		this.dc = dc;
		this.source = source;
	}
	
	/**
		Get the DbControl to use for database access.
	*/
	protected DbControl getDbControl()
	{
		return dc;
	}
	/**
		Get the source bioassay set.
	*/
	protected BioAssaySet getSource()
	{
		return source;
	}
	
	/**
		Set a snapshot manager that should be used with annotation-based
		fields. If no snapshot manager is provided, a new manager will 
		automatically be created if needed.
		@param snapshotManager The snapshot manager to use
	*/
	public void setSnapshotManager(SnapshotManager snapshotManager)
	{
		this.snapshotManager = snapshotManager;
	}
	
	/**
		Get the snapshot manager to use with annotation-based fields.
		If no snapshot manager has been configured a new one is automatically
		created.
		@return A snapshot manager object
	*/
	public SnapshotManager getSnapshotManager()
	{
		if (snapshotManager == null) snapshotManager = new SnapshotManager();
		return snapshotManager;
	}
	
	/**
		Get a static assay property field, such as name, description or id.
		Property lookup is not case-sensitive. Eg. both 'name' and 'Name' will
		match.
		
		@param name The property name
		@param exportTitle The title as it should be in the exported file
		@return An AssayPropertyField or null if not found
	*/
	protected AssayPropertyField getAssayPropertyField(String name, String exportTitle)
	{
		AssayPropertyField fld = null;
		Metadata<BioAssay> metadata = Metadata.getInstance(BioAssay.class);
		try
		{
			PropertyPath<BioAssay, ?> path = metadata.getPropertyPath(name, true);
			Type type = Type.fromHibernateType(path.getHibernateType());
			if (type == null) 
			{
				// The output is written to a text file, so if we don't know 
				// what it is, it will always end up as a string
				type = Type.STRING;
			}
			fld = ExportableFieldFactory.assayProperty(path.getPath(), exportTitle, type, null);
		}
		catch (RuntimeException ex)
		{}
		return fld;
	}
	
	/**
		Get an experimental factor assay field. A database query is used to 
		locate an experimental factor for the experiment that has the
		given name or external id. Depending on the underlying database, 
		the lookup may be case-sensitive or case-insensitive. If more than 
		one experimental factor is found an exception is thrown.
		
		@param name The name or external id of the experimental factor to find
		@param exportTitle The title as it should be in the exported file or null
			to use the name of the experimental factor
		@return An AnnotationAssayField or null if no experimental factor is found
	*/
	protected AnnotationAssayField getExperimentalFactorField(String name, String exportTitle)
	{
		if (factorQuery == null)
		{
			factorQuery = getSource().getExperiment().getExperimentalFactors();
			factorQuery.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			//WHERE name = :name OR externalId = :name
			Expression nameParameter = Expressions.parameter("name");
			factorQuery.restrictPermanent(
					Restrictions.or(
						Restrictions.eq(Hql.property("name"), nameParameter),
						Restrictions.eq(Hql.property("externalId"), nameParameter)
					)
			);
		}
		factorQuery.setParameter("name", name, Type.STRING);
		List<AnnotationType> result = factorQuery.list(dc);
		int size = result.size();
		AnnotationAssayField fld = null;
		if (size > 1)
		{
			throw new BaseException("Found " + result.size() + 
				" experimental factors for name/external id: " + name);
		}
		else if (size == 1)
		{
			AnnotationType at = result.get(0);
			if (exportTitle == null) exportTitle = at.getName();
			fld = ExportableFieldFactory.annotation(at, exportTitle, null);
			fld.setSnapshotManager(getSnapshotManager());
		}
		return fld;
	}
	
	/**
		Get an assay annotation field. A database query is used to locate an annotation 
		type that has been used on at least one assay in the bioassay set.
		Depending on the underlying database, the lookup may be case-sensitive or 
		case-insensitive. If more than one annotation type is found an exception is thrown.
		
		@param name The name or external id of the annotation type to find
		@param exportTitle The title as it should be in the exported file or null
			to use the name of the annotation type
		@return An AnnotationAssayField or null if no annotation type is found
	*/
	protected AnnotationAssayField getAssayAnnotationField(String name, String exportTitle)
	{
		if (assayAnnotationQuery == null)
		{
			assayAnnotationQuery = AnnotationType.getQuery(null);
			assayAnnotationQuery.setDistinct(true);
			assayAnnotationQuery.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			// WHERE name = :name or externalId = :name
			Expression nameParameter = Expressions.parameter("name");
			assayAnnotationQuery.restrictPermanent(
				Restrictions.or(
					Restrictions.eq(Hql.property("name"), nameParameter),
					Restrictions.eq(Hql.property("externalId"), nameParameter)
				)
			);
			// INNER JOIN annotations a WHERE a.annotationSet.id IN (:ids)
			assayAnnotationQuery.join(Hql.innerJoin("annotations", "a"));
			assayAnnotationQuery.restrictPermanent(
				Restrictions.in(
					Hql.property("a", "annotationSet.id"), 
					Expressions.parameter("assayAnnotationSets")
				)
			);
			List<BioAssay> assays = getSource().getBioAssays().list(dc);
			List<Integer> ids = new ArrayList<Integer>(assays.size());
			for (BioAssay ba : assays)
			{
				if (ba.isAnnotated()) ids.add(ba.getAnnotationSet().getId());
			}
			if (ids.size() == 0) ids.add(0);
			assayAnnotationQuery.setPermanentParameter("assayAnnotationSets", ids, Type.INT);
		}
		assayAnnotationQuery.setParameter("name", name, Type.STRING);
		List<AnnotationType> result = assayAnnotationQuery.list(dc);
		int size = result.size();
		AnnotationAssayField fld = null;
		if (size > 1)
		{
			throw new BaseException("Found " + result.size() + 
				" assay annotations for name/external id: " + name);
		}
		else if (size == 1)
		{
			AnnotationType at = result.get(0);
			if (exportTitle == null) exportTitle = at.getName();
			fld = ExportableFieldFactory.annotation(at, exportTitle, null);
			fld.setSnapshotManager(getSnapshotManager());
		}
		return fld;
	}
	
	/**
		Get a static or extended reporter property field, such as name, description 
		or id. Property lookup is not case-sensitive. Eg. both 'name' and 'Name' will
		match.
		
		@param name The name of the reporter property
		@param exportTitle The exported title or null to use the title
			of the extended property (which can be different from the name)
		@return A dynamic field, or null if no reporter property is found
	*/
	protected DynamicField getReporterProperty(String name, String exportTitle)
	{
		DynamicField fld = null;
		// Check the extended properties first
		List<ExtendedProperty> extended = ExtendedProperties.getProperties("ReporterData");
		if (extended != null)
		{
			for (ExtendedProperty ep : extended)
			{
				if (ep.getName().equalsIgnoreCase(name) || ep.getTitle().equalsIgnoreCase(name))
				{
					fld = ExportableFieldFactory.reporter(ep, exportTitle, null);
					break;
				}
			}
		}
		if (fld == null)
		{
			// If not found, use metadata
			Metadata<ReporterData> metadata = Metadata.getInstance(ReporterData.class);
			try
			{
				PropertyPath<ReporterData, ?> path = metadata.getPropertyPath(name, true);
				Type type = Type.fromHibernateType(path.getHibernateType());
				if (type == null) 
				{
					// The output is written to a text file, so if we don't know 
					// what it is, it will always end up as a string
					type = Type.STRING;
				}
				fld = ExportableFieldFactory.reporter(path.getPath(), exportTitle, type, Formula.AverageMethod.NONE, null);
			}
			catch (RuntimeException ex)
			{}
		}
		if (fld == null)
		{
			// Special cases to make everything compatible with the importer which
			// requires that either "External ID" or "Internal ID" columns are present
			// "Position" is also handled as a special case
			if ("External ID".equalsIgnoreCase(name))
			{
				fld = ExportableFieldFactory.reporter("externalId", name, Type.STRING, null, null);
			}
			else if ("Internal ID".equalsIgnoreCase(name))
			{
				fld = ExportableFieldFactory.reporter("id", name, Type.STRING, null, null);
			}
			else if ("Position".equals(name))
			{
				fld = ExportableFieldFactory.jep("pos()", name, Type.INT, null, null);
			}
		}
		return fld;
	}

	/**
		Find an extended raw data property with the given name and create
		a dynamic field from it using exportTitle as the title.
		@param name The name of the extended raw data property
		@param exportTitle The exported title or null to use the title
			of the extended property (which can be different from the name)
		@return A dynamic field, or null if no raw data property is found
	*/
	protected DynamicField getRawDataProperty(String name, String exportTitle)
	{
		DynamicField fld = null;
		// Check the extended properties first
		List<RawDataProperty> properties = getSource().getRawDataType().getProperties();
		if (properties != null)
		{
			for (RawDataProperty rp : properties)
			{
				if (rp.getName().equalsIgnoreCase(name) || rp.getTitle().equalsIgnoreCase(name))
				{
					fld = ExportableFieldFactory.rawData(rp, exportTitle, null);
					break;
				}
			}
		}
		if (fld == null)
		{
			// If not found, use metadata
			Metadata<RawData> metadata = Metadata.getInstance(RawData.class, getSource().getRawDataType().getEntityName());
			try
			{
				PropertyPath<RawData, ?> path = metadata.getPropertyPath(name, true);
				Type type = Type.fromHibernateType(path.getHibernateType());
				if (type == null) 
				{
					// The output is written to a text file, so if we don't know 
					// what it is, it will always end up as a string
					type = Type.STRING;
				}
				fld = ExportableFieldFactory.rawData(path.getPath(), exportTitle, type, Formula.AverageMethod.NONE, null);
			}
			catch (RuntimeException ex)
			{}
		}

		return fld;
	}

	
	/**
		Get an extra value dynamic field. A database query is used to 
		locate an extra value for the bioassay set that has the
		given name or external id. Depending on the underlying database, 
		the lookup may be case-sensitive or case-insensitive. If more than 
		one extra value is found an exception is thrown.
		
		@param name The name or external id of the experimental factor to find
		@param type The coordinate type of the extra value (eg. POSITION for reporter
			extra values, and SPOT for spot extra values)
		@param exportTitle The title as it should be in the exported file or null
			to use the name of the extra value
		@return A DynamicField or null if no extra value is found
	*/
	protected DynamicField getExtraValueField(String name, CoordinateType type, String exportTitle)
	{
		if (extraQuery == null)
		{
			extraQuery = getSource().getExtraValues();
			// INNER JOIN extraValueType evt
			extraQuery.joinPermanent(Hql.innerJoin("extraValueType", "evt"));
			// WHERE evt.name = :name OR evt.externalId = :name
			Expression nameParameter = Expressions.parameter("name");
			extraQuery.restrictPermanent(
				Restrictions.or(
					Restrictions.eq(Hql.property("evt", "name"), nameParameter),
					Restrictions.eq(Hql.property("evt", "externalId"), nameParameter)
				)
			);
			// WHERE dataCubeExtraValue.coordinateType = :type
			extraQuery.restrictPermanent(
					Restrictions.eq(
						Hql.property("dataCubeExtraValue.coordinateType"), 
						Expressions.parameter("type"))
				);
		}
		extraQuery.setParameter("name", name, Type.STRING);
		extraQuery.setParameter("type", type.getValue(), Type.INT);
		List<ExtraValue> result = extraQuery.list(dc);
		int size = result.size();
		DynamicField fld = null;
		if (size > 1)
		{
			throw new BaseException("Found " + result.size() + 
				" extra values with name/external id: " + name);
		}
		else if (size == 1)
		{
			ExtraValueType evt = result.get(0).getExtraValueType();
			if (exportTitle == null) exportTitle = evt.getName();
			fld = ExportableFieldFactory.extraValue(evt, exportTitle, null);
		}
		return fld;
	}
	
	/**
		Get a formula dynamic field. A database query is used to 
		locate the formula which must be a {@link net.sf.basedb.core.Formula.Type#COLUMN_EXPRESSION} 
		and be suitable for the  {@link RawDataType} and {@link IntensityTransform} 
		used by the source bioassay set. Depending on the underlying database, 
		the lookup may be case-sensitive or case-insensitive. If more than 
		one formula is found an exception is thrown.
		
		@param name The name of the formula to find
		@param exportTitle The title as it should be in the exported file or null
			to use the name of the formula
		@return A DynamicField or null if no extra value is found
	*/
	protected DynamicField getFormulaField(String name, String exportTitle)
	{
		if (formulaQuery == null)
		{
			BioAssaySet src = getSource();
			formulaQuery = Formula.getQuery(Formula.Type.COLUMN_EXPRESSION, src.getRawDataType(), src.getIntensityTransform());
			formulaQuery.include(Include.MINE, Include.SHARED, Include.IN_PROJECT, Include.OTHERS);
			formulaQuery.restrictPermanent(Restrictions.eq(Hql.property("name"), Expressions.parameter("name")));
		}
		formulaQuery.setParameter("name", name, Type.STRING);
		List<Formula> result = formulaQuery.list(dc);
		int size = result.size();
		DynamicField fld = null;
		if (size > 1)
		{
			throw new BaseException("Found " + result.size() + 
				" formulas with name: " + name);
		}
		else if (size == 1)
		{
			Formula f = result.get(0);
			if (exportTitle == null) exportTitle = f.getName();
			fld = ExportableFieldFactory.formula(f, exportTitle, null);
		}
		return fld;
	}
	
	/**
		Create a dynamic field for spot intensity if the field name has
		the pattern 'Ch X' where 'X' is a channel number between 1
		and the number of channels in the raw data type used by the 
		source bioassay set.
		@param fieldName The field name
		@return A dynamic field, or null if the name doesn't match the pattern
	*/
	protected DynamicField getIntensityField(String fieldName, String exportTitle)
	{
		DynamicField df = null;
		for (int i = 1; i <= getSource().getRawDataType().getChannels(); ++i)
		{
			if (fieldName.equalsIgnoreCase("Ch "+i))
			{
				df = ExportableFieldFactory.channel(i, exportTitle, null);
				break;
			}
		}
		return df;
	}

	
}
