/**
	$Id: ExtractLoader.java 5500 2010-11-18 07:26:30Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Sample;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.OverviewUtil;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for extracts. The forward-loading
	direction goes from {@link Sample}:s -&gt; {@link Extract}:s -&gt; 
	{@link LabeledExtract}:s. The reverse-loading direction is the opposite 
	direction. In both cases we also need to consider pooled extracts.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2010-11-18 08:26:30 +0100 (Thu, 18 Nov 2010) $
*/
public class ExtractLoader
	extends BasicItemNodeLoader<Extract>
{

	public ExtractLoader()
	{
		super(Item.EXTRACT, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<Extract>("extract", "Extract"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading extract nodes from a given parent sample
		of extract. In both cases the returned node is a folder-type node
		that contains the child extracts.
		@return The folder node for the extracts
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node returnNode = null;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.SAMPLE)
		{
			returnNode = createForwardNode((Sample)parentNode.getItem(dc), dc, context, parentNode);
		}
		else if (parentType == Item.EXTRACT)
		{
			returnNode = createForwardNode((Extract)parentNode.getItem(dc), dc, context, parentNode);
		}
		return returnNode;
	}
	
	/**
		Create reverse-loading extract nodes from a labeled extract node
		or from a pooled extract. In the first case, the returned node is an
		item-type node with the parent extract. In the second case, the returned
		node is a folder-type node with the pooled extracts. 
		@return A folder-type or item-type node
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node returnNode = null;
		Item parentItem = parentNode.getItemType();
		if (parentItem == Item.LABELEDEXTRACT)
		{
			returnNode = createReverseNode((LabeledExtract)parentNode.getItem(dc), dc, context, parentNode);
		}
		else if (parentItem == Item.EXTRACT)
		{
			returnNode = createReverseNode((Extract)parentNode.getItem(dc), dc, context, parentNode);
		}
		return returnNode;
	}
	// --------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of an extract. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node extractNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, extractNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, extractNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, extractNode);
	}

	/**
		Loads the pooled extracts and labled extracts that has been created from
		this extract.
		@see ExtractLoader#createForwardNode(DbControl, OverviewContext, Node)
		@see ExtractLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node extractNode)
	{
		getNodeLoader(context, Item.LABELEDEXTRACT).createForwardNode(dc, context, extractNode);
		getNodeLoader(context, Item.EXTRACT).createForwardNode(dc, context, extractNode);
	}
	
	/**
		If the extract is a pooled extract, loads the parent extracts that
		it was pooled from. Otherwise load the parent sample.
		@see ExtractLoader#createReverseNode(DbControl, OverviewContext, Node)
		@see SampleLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node extractNode)
	{
		Extract extract = (Extract)extractNode.getItem();
		if (extract.isPooled())
		{
			getNodeLoader(context, Item.EXTRACT).createReverseNode(dc, context, extractNode);
		}
		else
		{
			getNodeLoader(context, Item.SAMPLE).createReverseNode(dc, context, extractNode);
		}
	}
	// ---------------------------------------
	/**
		Create reverse-loading extract nodes of a pooled extract. Eg. the child nodes
		are the source extracts that was pooled to create the given product.
	*/
	@SuppressWarnings("unchecked")
	private Node createReverseNode(Extract product, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<Extract> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<Extract> query = (ItemQuery<Extract>)context.initQuery(
				product.getCreationEvent().getSources(), "name");
		ItemResultIterator<Extract> it = query.iterate(dc);
		int numExtracts = 0;
		while (it.hasNext())
		{
			numExtracts++;
			Extract source = it.next();
			// We need to handle circular references inline because
			// we can never require that a validator implementation handles this
			Node circularNode = OverviewUtil.getCircularRef(parentNode, source);
			if (circularNode != null)
			{
				context.createFailure(Validator.CIRCULAR_REFERENCE, parentNode, 
						"Circular reference to '" + source.getName() + "'", 
						new Fix("Modify parent extracts of '" + product.getName() + "'", product),
						new Fix("Modify parent extracts of '" + source.getName() + "'", source)
					);
			}
			else
			{
				if (folderNode == null)
				{
					folderNode = new Node("pooled", "Parent extracts", parentNode, ChildNodeDirection.REVERSE);
				}
				createItemNode(nf, source, source, false, folderNode, ChildNodeDirection.REVERSE);
			}
		}
		postValidateFolder(nf, folderNode, parentNode, numExtracts == 0);
		return folderNode;
	}

	/**
		Create a reverse-loading extract node from a labeled extract. Eg. the child node
		is the extract that is the parent of the (non-pooled) labeled extract.
	*/
	private Node createReverseNode(LabeledExtract labeledExtract, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<Extract> nf = getNodeFactory(dc, context);
		Extract extract = null;
		boolean denied = false;
		try
		{
			extract = labeledExtract.getExtract();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node extractNode = createItemNode(nf, extract, extract, denied, 
				parentNode, ChildNodeDirection.REVERSE);
		return extractNode;
	}

	/**
		Create forward-loading extract nodes for the extracts that has a given
		sample as their parent.
	*/
	private Node createForwardNode(Sample sample, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<Extract> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<Extract> query = context.initQuery(sample.getExtracts(), "name");
		ItemResultIterator<Extract> it = query.iterate(dc);
		while (it.hasNext())
		{
			Extract extract = it.next();
			if (folderNode == null)
			{
				folderNode = new Node("extracts", "Child extracts", parentNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, extract, extract, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, parentNode, false);
		return folderNode;
	}
	
	/**
		Create forward-loading extract nodes for the (pooled) extract that has a given
		extract among their parents.
	*/
	private Node createForwardNode(Extract parentExtract, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<Extract> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<Extract> query = context.initQuery(parentExtract.getChildExtracts(), "name");
		ItemResultIterator<Extract> it = query.iterate(dc);
		while (it.hasNext())
		{
			Extract childExtract = it.next();
			// We need to handle circular references inline because
			// we can never require that a validator implementation handles this
			Node circularNode = OverviewUtil.getCircularRef(parentNode, childExtract);
			if (circularNode != null)
			{
				context.createFailure(Validator.CIRCULAR_REFERENCE, parentNode, 
						"Circular reference to '" + childExtract.getName() + "'", 
						new Fix("Modify parent extracts of '" + childExtract.getName() + "'", childExtract),
						new Fix("Modify parent extracts of '" + parentExtract.getName() + "'", parentExtract)
						);
			}
			else
			{
				if (folderNode == null)
				{
					folderNode = new Node("extracts", "Child extracts", parentNode, ChildNodeDirection.FORWARD);
				}
				createItemNode(nf, childExtract, childExtract, false, folderNode, ChildNodeDirection.FORWARD);
			}
		}
		postValidateFolder(nf, folderNode, parentNode, false);
		return folderNode;
	}

}
