/**
	$Id: HybridizationLoader.java 5500 2010-11-18 07:26:30Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.Scan;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.cache.IndexedCacheKey;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for hybridizations. There are two forward-loading
	directions:
	<ol>
	<li>{@link LabeledExtract}:s -&gt; {@link Hybridization}:s -&gt; 
	{@link Scan}:s. 
	<li>{@link ArraySlide}:s -&gt; {@link Hybridization}:s -&gt; 
	{@link Scan}:s
	</ol>
	
	The reverse-loading direction starts with a scan and moves to both
	labeled extracts and array slide.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2010-11-18 08:26:30 +0100 (Thu, 18 Nov 2010) $
*/
public class HybridizationLoader
	extends BasicItemNodeLoader<Hybridization>
{
	public HybridizationLoader()
	{
		super(Item.HYBRIDIZATION, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<Hybridization>("hybridization", "Hybridization"));
	}

	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading hybridization nodes from either a labeled extract or
		an array slide node. The returned node is either folder-type node containing 
		nodes for each hybridization that has been created from the labeled extract
		or an item node with the hybridization on the array slide.
		@return The folder node with the hybridizations, or the item node (can be null)
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node returnNode = null;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.LABELEDEXTRACT)
		{
			returnNode = createForwardNode((LabeledExtract)parentNode.getItem(dc), dc, context, parentNode);
		}
		else if (parentType == Item.ARRAYSLIDE)
		{
			returnNode = createForwardNode((ArraySlide)parentNode.getItem(dc), dc, context, parentNode);
		}
		return returnNode;
	}
	
	/**
		Create a reverse-loading hybridization node from the scan node.
		@return A hybridization node, or null if the scan doesn't have a 
			hybridization
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node scanNode)
	{
		NodeFactory<Hybridization> nf = getNodeFactory(dc, context);
		Hybridization hyb = null;
		boolean denied = false;
		int arrayNum = 0;
		try
		{
			Scan scan = (Scan)scanNode.getItem(dc);
			hyb = scan.getHybridization();
			Node rbaNode = scanNode.getFirstParent(new ItemTypeFilter(Item.RAWBIOASSAY));
			if (rbaNode != null)
			{
				RawBioAssay rba = (RawBioAssay)rbaNode.getItem();
				arrayNum = rba.getArrayNum();
			}
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Object cacheKey = new IndexedCacheKey(hyb, arrayNum);
		Node hybNode = createItemNode(nf, hyb, cacheKey, denied, scanNode, ChildNodeDirection.REVERSE);
		return hybNode;
	}
	// --------------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of a hybridization. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Hyb. station: {@link HardwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node hybNode)
	{
		getNodeLoader(context, Item.ANNOTATIONTYPE).createPropertyNode(dc, context, hybNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, hybNode);
		getNodeLoader(context, Item.HARDWARE).createPropertyNode(dc, context, hybNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, hybNode);
	}
	
	/**
		Load forward-loading child nodes:
		<ul>
		<li>Scans: {@link ScanLoader#createForwardNode(DbControl, OverviewContext, Node)}
		<li>Array slide in reverse direction (if the parent node is a labeled extract):
			{@link ArraySlideLoader#createReverseNode(DbControl, OverviewContext, Node)}
		<li>Labeled extracts in reverse direction (if the parent node is the array slide):
			{@link LabeledExtractLoader#createReverseNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node hybNode)
	{
		getNodeLoader(context, Item.SCAN).createForwardNode(dc, context, hybNode);
		Node parent = hybNode.getParent();
		if (parent != null && parent.getNodeType() == Node.Type.FOLDER) 
		{
			parent = parent.getParent();
		}
		Item parentItemType = parent == null ? null : parent.getItemType();
		if (parentItemType == Item.LABELEDEXTRACT)
		{
			getNodeLoader(context, Item.ARRAYSLIDE).createReverseNode(dc, context, hybNode);
		}
		if (parentItemType == Item.ARRAYSLIDE)
		{
			getNodeLoader(context, Item.LABELEDEXTRACT).createReverseNode(dc, context, hybNode);
		}
	}
	
	/**
		Load reverse child nodes:
		<ul>
		<li>Labeled extracts: {@link LabeledExtractLoader#createReverseNode(DbControl, OverviewContext, Node)}
		<li>Array slide: {@link ArraySlideLoader#createReverseNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node hybNode)
	{
		getNodeLoader(context, Item.LABELEDEXTRACT).createReverseNode(dc, context, hybNode);
		getNodeLoader(context, Item.ARRAYSLIDE).createReverseNode(dc, context, hybNode);
	}
	// --------------------------------------
	
	/**
		Create forward-loading hybridization nodes for the hybridizations 
		that has a given labeled extract as a source.
	*/
	private Node createForwardNode(LabeledExtract extract, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<Hybridization> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<Hybridization> query = context.initQuery(extract.getHybridizations(), "name");
		ItemResultIterator<Hybridization> it = query.iterate(dc);
		while (it.hasNext())
		{
			Hybridization hyb = it.next();
			int arrayNum = hyb.getCreationEvent().getSourceGroup(extract);
			Object cacheKey = arrayNum == 0 ? hyb : new IndexedCacheKey(hyb, arrayNum);
			if (folderNode == null)
			{
				folderNode = new Node("hybridizations", "Hybridizations", parentNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, hyb, cacheKey, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, parentNode, false);
		return folderNode;
	}

	/**
		Create a forward-loading hybridization node from an array slide.
	*/
	private Node createForwardNode(ArraySlide slide, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<Hybridization> nf = getNodeFactory(dc, context);
		Hybridization hyb = null;
		boolean denied = false;
		try
		{
			hyb = slide.getHybridization();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node hybNode = createItemNode(nf, hyb, hyb, denied, parentNode, ChildNodeDirection.FORWARD);
		return hybNode;
	}

}
