/**
	$Id: LabeledExtractLoader.java 5518 2010-11-22 08:02:13Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Extract;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.OverviewUtil;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.LabeledExtractNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for labeled extracts. The forward-loading
	direction goes from {@link Extract}:s -&gt; {@link LabeledExtract}:s -&gt; 
	{@link Hybridization}:s. The reverse-loading direction is the opposite 
	direction. In both cases we also need to consider pooled labeled extracts.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2010-11-22 09:02:13 +0100 (Mon, 22 Nov 2010) $
*/
public class LabeledExtractLoader
	extends BasicItemNodeLoader<LabeledExtract>
{

	public LabeledExtractLoader()
	{
		super(Item.LABELEDEXTRACT, ALLOW_ROOT_NODE, 
				new LabeledExtractNameGenerator("labeledextract", "Labeled extract"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading labeled extract nodes from a given parent extract
		or labeled extract node. In both cases the returned node is a folder-type node
		that contains the child labeled extracts.
		@return The folder node for the labeled extracts
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node folderNode = null;
		Item parentItem = parentNode.getItemType();
		if (parentItem == Item.EXTRACT)
		{
			folderNode = createForwardNode((Extract)parentNode.getItem(dc), dc, context, parentNode);
		}
		else if (parentItem == Item.LABELEDEXTRACT)
		{
			folderNode = createForwardNode((LabeledExtract)parentNode.getItem(dc), dc, context, parentNode);
		}
		return folderNode;
	}
	
	/**
		Create reverse-loading labeled extract nodes from a hybridization node
		or from a pooled labeled extract. 
		@return A folder node with the labeled extracts
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node parentNode)
	{
		Node folderNode = null;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.LABELEDEXTRACT)
		{
			folderNode = createReverseNode((LabeledExtract)parentNode.getItem(dc), dc, context, parentNode);
		}
		else if (parentType == Item.HYBRIDIZATION)
		{
			folderNode = createReverseNode((Hybridization)parentNode.getItem(dc), dc, context, parentNode);
		}
		return folderNode;
	}
	// --------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of a labeled extract. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node extractNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, extractNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, extractNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, extractNode);
	}

	/**
		Loads the hybridizations and labeled extracts that has been created from
		this labeled extract.
		@see HybridizationLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node extractNode)
	{
		getNodeLoader(context, Item.HYBRIDIZATION).createForwardNode(dc, context, extractNode);
		getNodeLoader(context, Item.LABELEDEXTRACT).createForwardNode(dc, context, extractNode);
	}
	
	/**
		If the labeled extract is a pooled labeled extract, loads the labeled extracts
		it was pooled from. Otherwise load the parent extract.
		@see LabeledExtractLoader#createReverseNode(DbControl, OverviewContext, Node)
		@see ExtractLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node extractNode)
	{
		LabeledExtract extract = (LabeledExtract)extractNode.getItem();
		if (extract.isPooled())
		{
			getNodeLoader(context, Item.LABELEDEXTRACT).createReverseNode(dc, context, extractNode);
		}
		else
		{
			getNodeLoader(context, Item.EXTRACT).createReverseNode(dc, context, extractNode);
		}
	}
	// ---------------------------------------
	
	/**
		Create reverse-loading labeled extract nodes of a pooled labeled extract. Eg. 
		the child nodes are the source labeled extracts that was pooled to create the 
		given product.
	*/
	@SuppressWarnings("unchecked")
	private Node createReverseNode(LabeledExtract product, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<LabeledExtract> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<LabeledExtract> query = (ItemQuery<LabeledExtract>)context.initQuery(
				product.getCreationEvent().getSources(), "name");
		ItemResultIterator<LabeledExtract> it = query.iterate(dc);
		int numExtracts = 0;
		while (it.hasNext())
		{
			numExtracts++;
			LabeledExtract source = it.next();
			// We need to handle circular references inline because
			// we can never require that a validator implementation handles this
			Node circularNode = OverviewUtil.getCircularRef(parentNode, source);
			if (circularNode != null)
			{
				context.createFailure(Validator.CIRCULAR_REFERENCE, parentNode, 
						"Circular reference to '" + source.getName() + "'", 
						new Fix("Modify parent labeled extracts of '" + product.getName() + "'", product),
						new Fix("Modify parent labeled extracts of '" + source.getName() + "'", source)
					);
			}
			else
			{
				if (folderNode == null)
				{
					folderNode = new Node("pooled", "Parent labeled extracts", parentNode, ChildNodeDirection.REVERSE);
				}
				createItemNode(nf, source, source, false, folderNode, ChildNodeDirection.REVERSE);
			}
		}
		postValidateFolder(nf, folderNode, parentNode, numExtracts == 0);
		return folderNode;
	}
	
	/**
		Create reverse-loading labeled extract nodes from a hybridization. Eg. the child nodes
		are the labeled extracts that has been used on a hybridization.
	*/
	private Node createReverseNode(Hybridization hyb, DbControl dc, OverviewContext context, Node parentNode)
	{
		Node folderNode = null;
		NodeFactory<LabeledExtract> nf = getNodeFactory(dc, context);

		// If we have followed this path from a raw bioassay we should
		// only load labeled extracts on the specific 'array index'
		Node rbaNode = parentNode.getFirstParent(new ItemTypeFilter(Item.RAWBIOASSAY));
		int arrayNum = 0;
		if (rbaNode != null)
		{
			arrayNum = ((RawBioAssay)rbaNode.getItem()).getArrayNum();
		}
		ItemQuery<LabeledExtract> query = context.initQuery(hyb.getLabeledExtracts(arrayNum), "name");
		ItemResultIterator<LabeledExtract> it = query.iterate(dc);
		while (it.hasNext())
		{
			LabeledExtract extract = it.next();
			if (folderNode == null)
			{
				folderNode = new Node("labeledextracts", "Labeled extracts", parentNode, ChildNodeDirection.REVERSE);
			}
			createItemNode(nf, extract, extract, false, folderNode, ChildNodeDirection.REVERSE);
		}
		postValidateFolder(nf, folderNode, parentNode, rbaNode == null);
		return folderNode;
	}

	/**
		Create forward-loading labeled extract nodes for the labeled extracts that 
		has a given extract as their parent.
	*/
	private Node createForwardNode(Extract extract, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<LabeledExtract> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<LabeledExtract> query = context.initQuery(extract.getLabeledExtracts(), "name");
		ItemResultIterator<LabeledExtract> it = query.iterate(dc);
		while (it.hasNext())
		{
			LabeledExtract childExtract = it.next();
			if (folderNode == null)
			{
				folderNode = new Node("labeledextracts", "Child labeled extracts", parentNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, childExtract, childExtract, false, folderNode, ChildNodeDirection.FORWARD);
		}		
		postValidateFolder(nf, folderNode, parentNode, false);
		return folderNode;
	}
	
	/**
		Create forward-loading labeled extract nodes for the (pooled) labeled extract that 
		has a given labeled extract among their parents.
	*/
	private Node createForwardNode(LabeledExtract parentExtract, DbControl dc, OverviewContext context, Node parentNode)
	{
		NodeFactory<LabeledExtract> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<LabeledExtract> query = context.initQuery(parentExtract.getChildLabeledExtracts(), "name");;
		ItemResultIterator<LabeledExtract> it = query.iterate(dc);
		while (it.hasNext())
		{
			LabeledExtract childExtract = it.next();
			// We need to handle circular references inline because
			// we can never require that a validator implementation handles this
			Node circularNode = OverviewUtil.getCircularRef(parentNode, childExtract);
			if (circularNode != null)
			{
				context.createFailure(Validator.CIRCULAR_REFERENCE, parentNode, 
						"Circular reference to '" + childExtract.getName() + "'", 
						new Fix("Modify parent labeled extracts of '" + childExtract.getName() + "'", childExtract),
						new Fix("Modify parent labeled extracts of '" + parentExtract.getName() + "'", parentExtract)
					);
			}
			else
			{
				if (folderNode == null)
				{
					folderNode = new Node("labeledextracts", "Child labeled extracts", parentNode, ChildNodeDirection.FORWARD);
				}
				createItemNode(nf, childExtract, childExtract, false, folderNode, ChildNodeDirection.FORWARD);
			}
		}
		postValidateFolder(nf, folderNode, parentNode, false);
		return folderNode;
	}
	
}
