/**
	$Id: RawBioAssayLoader.java 5500 2010-11-18 07:26:30Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;


import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.Scan;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for raw bioassays. The forward-loading
	direction goes from {@link Scan}:s -&gt; {@link RawBioAssay}:s -&gt; 
	{@link Experiment}:s. The reverse-loading direction is the opposite 
	direction.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2010-11-18 08:26:30 +0100 (Thu, 18 Nov 2010) $
*/
public class RawBioAssayLoader
	extends BasicItemNodeLoader<RawBioAssay>
{
	public RawBioAssayLoader()
	{
		super(Item.RAWBIOASSAY, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<RawBioAssay>("rawbioassay", "Raw bioassay"));
	}

	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading raw bioassay nodes from a given scan node.
		The returned node is a folder-type node containing item nodes for 
		each raw bioassay that has been created from the scan.
		@return The folder node for the raw bioassays
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node scanNode)
	{
		NodeFactory<RawBioAssay> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		Scan scan = (Scan)scanNode.getItem(dc);
		ItemQuery<RawBioAssay> query = context.initQuery(scan.getRawBioAssays(), "name");
		
		// If we have a labeled extract as parent we need to know the array number
		// to be able to load the correct raw bioassays -- The same hyb may have
		// other raw bioassays that are NOT related to the same labeled extract
		int arrayNum = 0;
		Hybridization hyb = null;
		Node hybNode = scanNode.getFirstParent(new ItemTypeFilter(Item.HYBRIDIZATION));
		if (hybNode != null)
		{
			Node labeledExtractNode = hybNode.getFirstParent(new ItemTypeFilter(Item.LABELEDEXTRACT));
			if (labeledExtractNode != null)
			{
				hyb = (Hybridization)hybNode.getItem(dc);
				LabeledExtract extract = (LabeledExtract)labeledExtractNode.getItem();
				arrayNum = hyb.getCreationEvent().getSourceGroup(extract);
			}
		}
		if (arrayNum > 0)
		{
			query.restrict(Restrictions.eq(Hql.property("arrayNum"), Expressions.integer(arrayNum)));
		}
		
		ItemResultIterator<RawBioAssay> it = query.iterate(dc);
		while (it.hasNext())
		{
			RawBioAssay rba = it.next();
			if (folderNode == null)
			{
				folderNode =  new Node("rawbioassays", "Raw bioassays", scanNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, rba, rba, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, scanNode, false);
		return folderNode;
	}

	/**
		Create a reverse-loading raw bioassay nodes from a given experiment node.
		The returned node is a folder-type node containing nodes for each raw
		bioassay is part of the experiment.
		@return The folder node for the raw bioassays
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node experimentNode)
	{
		NodeFactory<RawBioAssay> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		Experiment experiment = (Experiment)experimentNode.getItem(dc);
		ItemQuery<RawBioAssay> query = context.initQuery(experiment.getRawBioAssays(), "name");
		ItemResultIterator<RawBioAssay> it = query.iterate(dc);
		while (it.hasNext())
		{
			RawBioAssay rba = it.next();
			if (folderNode == null)
			{
				folderNode =  new Node("rawbioassays", "Raw bioassays", experimentNode, ChildNodeDirection.REVERSE);
			}
			createItemNode(nf, rba, rba, false, folderNode, ChildNodeDirection.REVERSE);
		}
		postValidateFolder(nf, folderNode, experimentNode, false);
		return folderNode;
	}
	// ------------------------------
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Load property nodes for a raw bioassay:
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Data files: {@link DataFileLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Platform: {@link PlatformLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Software: {@link SoftwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Array design: {@link ArrayDesignLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.PLATFORM).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.FILESETMEMBER).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.ARRAYDESIGN).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.SOFTWARE).createPropertyNode(dc, context, rawBioAssayNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, rawBioAssayNode);
	}
	/**
		Loads all experiment nodes where the given raw bioassay is used.
		@see ExperimentLoader#createForwardNode(DbControl, OverviewContext, Node) 
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		getNodeLoader(context, Item.EXPERIMENT).createForwardNode(dc, context, rawBioAssayNode);
	}
	/**
		Loads the scan node that this raw bioassay is created from.
		@see ScanLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node rawBioAssayNode)
	{
		getNodeLoader(context, Item.SCAN).createReverseNode(dc, context, rawBioAssayNode);
	}
	// --------------------------------

}
