/**
	$Id: BasicItemNodeValidatorFactory.java 5500 2010-11-18 07:26:30Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.HashMap;
import java.util.Map;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioSource;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.util.ClassUtil;

/**
	Node validator factory implementation for {@link BasicItem} node validators.
	The key used in {@link #createNodeValidator(Object)} should normally be 
	an {@link Item} object.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2010-11-18 08:26:30 +0100 (Thu, 18 Nov 2010) $
*/
public class BasicItemNodeValidatorFactory
	implements NodeValidatorFactory<BasicItem, Object>
{
	
	private final Map<Object, Class<? extends NodeValidator<? extends BasicItem>>> validators;
	private boolean useNullValidator;
	private boolean disabled;
	
	/**
		Creates a new factory.
	*/
	public BasicItemNodeValidatorFactory()
	{
		validators = new HashMap<Object, Class<? extends NodeValidator<? extends BasicItem>>>();
		this.useNullValidator = true;
		this.disabled = false;
		registerDefaultNodeValidators();
	}
	
	/*
		From the NodeValidatorFactory interface
		---------------------------------------
	*/
	
	
	@Override
	public boolean isDisabled() 
	{
		return disabled;
	}

	@Override
	public void setDisabled(boolean disabled) 
	{
		this.disabled = disabled;
	}

	/**
		Create a new node validator that can validate nodes with items of 
		the specified item type. If no validator has been registered for the 
		given item type, a {@link NullNodeValidator} is returned or an exception 
		is thrown depending on the {@link #useNullValidatorIfNotFound()} option.
		
		@param key An {@link Item} object
		@throws ItemNotFoundException If no node loader is found and the 
			{@link #useNullValidatorIfNotFound()} setting is false
		@return A node validator object
	*/
	@Override
	public NodeValidator<? extends BasicItem> createNodeValidator(Object key)
	{
		NodeValidator<? extends BasicItem> validator = null;
		Class<? extends NodeValidator<? extends BasicItem>> validatorClass = 
			disabled ? null : validators.get(key);
		if (validatorClass == null)
		{
			if (useNullValidator || disabled)
			{
				validator = new NullNodeValidator<BasicItem>(true, false);
			}
			else
			{
				throw new ItemNotFoundException("Node validator for key: " + key);
			}
		}
		else
		{
			try
			{
				validator = validatorClass.newInstance();
			}
			catch (Exception ex)
			{
				throw new BaseException("Can't create node validator: " + validatorClass, ex);
			}
		}
		return validator;
	}
	// --------------------------------------
	
	/**
		Get the null validator setting.
		@see #setUsNullValidatorIfNotFound(boolean)
	*/
	public boolean useNullValidatorIfNotFound()
	{
		return useNullValidator;
	}
	
	/**
		Set a flag that determines if a {@link NullNodeValidator} should be used
		or if an {@link ItemNotFoundException} should be thrown
		if no registered node validator is found for a specific key.
		@param useNullValidator TRUE to use an {@link NullNodeValidator}, FALSE to throw
			an exception
	 */
	public void setUsNullValidatorIfNotFound(boolean useNullValidator)
	{
		this.useNullValidator = useNullValidator;
	}

	
	/**
		Register a node validator implementation. The implementation must have a
		public no-argument constructor. The given class is checked with
		{@link ClassUtil#checkAndLoadClass(ClassLoader, String, boolean, Class...)}
		
		@param key The item type the node validator handles
		@param validatorClass The class name of a class that implements {@link NodeValidator},
			or null to unregister an existig validator
		@throws NoSuchMethodException If the given class has no public no-argument
			constructor
		@throws ClassCastException If the given class doesn't implement the
			{@link NodeValidator} interface
	*/
	public void registerNodeValidator(Item key, Class<? extends NodeValidator<? extends BasicItem>> validatorClass)
		throws NoSuchMethodException, ClassCastException
	{
		if (validatorClass != null) ClassUtil.checkClass(validatorClass, true, NodeValidator.class);
		registerCheckedNodeValidator(key, validatorClass);
	}
	
	/**
		Register a node loader implementation that you are sure fulfills the
		requirements. Eg. it must have a public no-argument constructor and
		implement the {@link NodeValidator} interface.
		@see #registerNodeValidator(Item, Class)
	*/
	protected void registerCheckedNodeValidator(Object key, 
			Class<? extends NodeValidator<? extends BasicItem>> validatorClass)
	{
		if (validatorClass == null)
		{
			validators.remove(key);
		}
		else
		{
			validators.put(key, validatorClass);
		}
	}
	

	/**
		Registers default item node validators from {@link BioSource}
		to {@link Experiment}, etc.
	*/
	protected void registerDefaultNodeValidators()
	{
		// Main validators
		registerCheckedNodeValidator(Item.BIOSOURCE, BioSourceValidator.class);
		registerCheckedNodeValidator(Item.SAMPLE, SampleValidator.class);
		registerCheckedNodeValidator(Item.EXTRACT, ExtractValidator.class);
		registerCheckedNodeValidator(Item.LABELEDEXTRACT, LabeledExtractValidator.class);
		registerCheckedNodeValidator(Item.ARRAYDESIGN, ArrayDesignValidator.class);
		registerCheckedNodeValidator(Item.ARRAYBATCH, ArrayBatchValidator.class);
		registerCheckedNodeValidator(Item.ARRAYSLIDE, ArraySlideValidator.class);
		registerCheckedNodeValidator(Item.HYBRIDIZATION, HybridizationValidator.class);
		registerCheckedNodeValidator(Item.SCAN, ScanValidator.class);
		registerCheckedNodeValidator(Item.RAWBIOASSAY, RawBioAssayValidator.class);
		registerCheckedNodeValidator(Item.EXPERIMENT, ExperimentValidator.class);
//		
//		// Property node validators
		registerCheckedNodeValidator(Item.PROTOCOL, ProtocolValidator.class);
		registerCheckedNodeValidator("PROTOCOL.PARAMETER", ProtocolParameterValidator.class);
		registerCheckedNodeValidator("EXPERIMENTAL.FACTOR", ExperimentalFactorValidator.class);
		registerCheckedNodeValidator(Item.SOFTWARE, SoftwareValidator.class);
		registerCheckedNodeValidator(Item.HARDWARE, HardwareValidator.class);
		registerCheckedNodeValidator(Item.ANNOTATION, AnnotationValidator.class);
		registerCheckedNodeValidator(Item.PLATFORM, PlatformValidator.class);
		registerCheckedNodeValidator(Item.FILESETMEMBER, DataFileValidator.class);
		registerCheckedNodeValidator(Item.ANYTOANY, AnyToAnyValidator.class);
	}

}
