/**
	$Id: LabeledExtractValidator.java 5518 2010-11-22 08:02:13Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.BioMaterialEvent;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.Label;
import net.sf.basedb.core.LabeledExtract;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.util.overview.Failure;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for labeled extracts. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_LABELED_EXTRACT}
	<li>Denied access to label: {@link Validator#DENIED_LABEL}
	<li>Different label for pooled child: {@link Validator#DIFFERENT_LABEL}
	<li>Labeled extract has incorrect array index on hyb: {@link Validator#INCORRECT_ARRAY_NUM}
	<li>Duplicates of same label on hyb: {@link Validator#NONUNIQUE_LABEL}
	<li>Number of labeled extracts on hyb != number of channels in experiment
		{@link Validator#INCORRECT_NUM_LABELEDEXTRACTS}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2010-11-22 09:02:13 +0100 (Mon, 22 Nov 2010) $
*/
public class LabeledExtractValidator
	extends NameableNodeValidator<LabeledExtract>
{
	
	public LabeledExtractValidator()
	{
		super(Validator.MISSING_LABELED_EXTRACT, null);
	}
	
	/* 
		From BasicValidator class
		-------------------------
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);

		// Check if access is granted to the label
		try
		{
			((LabeledExtract)node.getItem(dc)).getLabel();
		}
		catch (PermissionDeniedException ex)
		{
			context.createFailure(Validator.DENIED_LABEL, node, null);
		}

		// Other validations depends on the type of the parent node
		if (parentNode == null) return; // We are the root node
		if (parentNode.getNodeType() == Node.Type.FOLDER) parentNode = parentNode.getParent();
		if (parentNode == null) return;
		Item parentType = parentNode.getItemType();
		if (parentType == Item.LABELEDEXTRACT)
		{
			validatePooledLabeledExtract(dc, context, node, parentNode);
		}
		else if (parentType == Item.HYBRIDIZATION)
		{
			validateLabeledExtractOnHyb(dc, context, node, parentNode);
		}
		// The third case is parentType == Item.EXTRACT but this currently doesn't 
		// need any special validation
	}
	
	@Override
	public void postValidateFolder(DbControl dc, OverviewContext context, Node folderNode, 
			Node parentNode)
	{
		super.postValidateFolder(dc, context, folderNode, parentNode);
		if (parentNode.getItemType() != Item.HYBRIDIZATION) return;

		Hybridization hyb = (Hybridization)parentNode.getItem(dc);
		BioMaterialEvent creationEvent = hyb.getCreationEvent();
		
		// Check that the combo 'array-index/label' is unique on the hyb
		Map<String, LabeledExtract> usedLabelIndexes = new HashMap<String, LabeledExtract>();
		Set<LabeledExtract> duplicates = new HashSet<LabeledExtract>();
		int minArrayIndex = Integer.MAX_VALUE;
		int maxArrayIndex = 0;
		int numExtracts = 0;
		if (folderNode != null)
		{
			for (Node leNode : folderNode.getChildren())
			{
				numExtracts++;
				LabeledExtract extract = (LabeledExtract)leNode.getItem(dc);
				try
				{
					Label label = extract.getLabel();
					int arrayIndex = creationEvent.getSourceGroup(extract);
					if (arrayIndex < minArrayIndex) minArrayIndex = arrayIndex;
					if (arrayIndex > maxArrayIndex) maxArrayIndex = arrayIndex;
					String labelIndex = label.getId() + ":" + arrayIndex;
					if (usedLabelIndexes.containsKey(labelIndex))
					{
						duplicates.add(extract);
						duplicates.add(usedLabelIndexes.get(labelIndex));
					}
					else
					{
						usedLabelIndexes.put(labelIndex, extract);
					}
				}
				catch (PermissionDeniedException ex)
				{}
			}
		}
		if (duplicates.size() > 0)
		{
			Failure failure = context.createFailure(Validator.NONUNIQUE_LABEL, parentNode, null);
			if (failure != null)
			{
				if (minArrayIndex == maxArrayIndex) failure.setDescriptor(maxArrayIndex);
				failure.addFix(new Fix("Change labeled extracts on hybridization", hyb));
				for (LabeledExtract le : duplicates)
				{
					failure.addFix(new Fix("Change label of '" + le.getName() + "'", le));
				}
			}
		}
		
		// Check that the number of labeled extracts per array-index matches 
		// the number of channels if we are descending from at least a raw bioassay
		Node rbaNode = parentNode.getFirstParent(new ItemTypeFilter(Item.RAWBIOASSAY));
		if (rbaNode != null)
		{
			RawBioAssay rba = (RawBioAssay)rbaNode.getItem();
			int numChannels = rba.getRawDataType().getChannels();
			if (numExtracts != numChannels)
			{
				Fix fix = numExtracts > numChannels ?
					new Fix("Remove labeled extracts from hybridization", hyb) :
					new Fix("Add labeled extracts to hybridization", hyb);
				String message = numExtracts > numChannels ?
					"Too many labeled extracts" : "Too few labeled extracts";
				message += ": " + numExtracts + " (" + numChannels + " expected)";
				Failure failure = context.createFailure(Validator.INCORRECT_NUM_LABELEDEXTRACTS, 
					parentNode, message, fix);
				if (failure != null)
				{
					if (minArrayIndex == maxArrayIndex) failure.setDescriptor(maxArrayIndex);
				}
			}

		}
	}

	/**
		The suggested fix is to add a source labeled extracts to the 
		either the hybridization or the pooled labeled extract.
	*/
	@Override
	protected Fix getMissingItemFix(Node parentNode)
	{
		Fix fix = null;
		Item parentType = parentNode.getItemType();
		BasicItem parentItem = parentNode.getItem();
		if (parentType == Item.HYBRIDIZATION)
		{
			fix =  new Fix("Add labeled extracts to '" + ((Hybridization)parentItem).getName() + "'", parentItem);
		}
		else if (parentType == Item.LABELEDEXTRACT)
		{
			fix =  new Fix("Add parent labeled extracts to '" + ((LabeledExtract)parentItem).getName() + "'", parentItem);
		}
		return fix;
	}
	// ----------------------------
	
	private void validateLabeledExtractOnHyb(DbControl dc, OverviewContext context, Node extractNode, Node hybNode)
	{
		// Check that the labeled extract's array index is within the expected range
		Hybridization hyb = (Hybridization)hybNode.getItem(dc);
		int numArrays = hyb.getNumArrays();
		LabeledExtract extract = (LabeledExtract)extractNode.getItem(dc);
		int arrayIndex = hyb.getCreationEvent().getSourceGroup(extract);
		if (arrayIndex > numArrays || arrayIndex < 1)
		{
			context.createFailure(Validator.INCORRECT_ARRAY_NUM, extractNode,
					"Incorrect sub-array index: " + arrayIndex + 
						" (expected 1.." + numArrays + ")",
					new Fix("Change array index for labeled extract", hyb), 
					new Fix("Increase arrays on hybridization", hyb));
		}
	}

	private void validatePooledLabeledExtract(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		// Check that the child has the same label as the parent
		LabeledExtract parent = (LabeledExtract)parentNode.getItem(dc);
		Label parentLabel = null;
		try
		{
			parentLabel = parent.getLabel();
		}
		catch (PermissionDeniedException ex)
		{}
		Label label = null;
		LabeledExtract child = (LabeledExtract)node.getItem(dc);
		try
		{
			label = child.getLabel();
		}
		catch (PermissionDeniedException ex)
		{}
		if ((parentLabel == null && label != null) || 
				(parentLabel != null && !parentLabel.equals(label)))
		{
			context.createFailure(Validator.DIFFERENT_LABEL, node, null,
				new Fix("Change label of '" + child.getName() + "'", child),
				new Fix("Change label of '" + parent.getName() + "'", parent)
				);
		}
	}
	
}
