/**
	$Id: JepExtraValueCalculator.java 5481 2010-11-08 13:49:35Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins;

import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.BioAssay;
import net.sf.basedb.core.BioAssaySet;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.DynamicSpotQuery;
import net.sf.basedb.core.ExtraValueType;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemParameterType;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Job;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.RequestInformation;
import net.sf.basedb.core.SpotExtraValueBatcher;
import net.sf.basedb.core.StringParameterType;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.VirtualColumn;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.AbstractAnalysisPlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.Permissions;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.signal.SignalHandler;
import net.sf.basedb.core.signal.SignalTarget;
import net.sf.basedb.core.signal.ThreadSignalHandler;
import net.sf.basedb.util.BioAssaySetUtil;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
	This plugin caluclates extra values from a bioassayset using mathematical
	formulas entered by the user. 

	@author nicklas
	@version 2.0
	@base.modified $Date: 2010-11-08 14:49:35 +0100 (Mon, 08 Nov 2010) $
*/
public class JepExtraValueCalculator
	extends AbstractAnalysisPlugin
	implements InteractivePlugin, SignalTarget
{

	/**
	   Plugin description
	*/
	private static final About about = 
		new AboutImpl
		(
			"JEP extra value calculator",
			"This plugin is used to calculate extra values for a bioassay set using "+
			"expressions entered by the user. The expressions are parsed with the JEP package.", 
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2006, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
		);
	
	private static final Set<Permissions> permissions = new HashSet<Permissions>();
	private static final Set<GuiContext> guiContexts = Collections.unmodifiableSet(
			new HashSet<GuiContext>(Arrays.asList(CONTEXT_BIOASSAYSET, CONTEXT_BIOASSAYS)));
	
	private static final PluginParameter<String> expressionParameter = new PluginParameter<String>(
			"expression", "Expression", "The expression to use for calculating the extra value. The expression " +
			"is parsed with the JEP package.", 
			new StringParameterType(255, null, true)
		);

	private ThreadSignalHandler signalHandler;
	
	private RequestInformation configureJob;
	
	private PluginParameter<ExtraValueType> extraValueTypeParameter;

	public JepExtraValueCalculator()
	{}
	
	/*
		From the Plugin interface
		-------------------------------------------
	*/
	public About getAbout()
	{
		return about;
	}
	public boolean supportsConfigurations()
	{
		return false;
	}
	public boolean requiresConfiguration()
	{
		return false;
	}
	/**
	 	Request use access to Experiment:s and ExtraValueType:s and read access to Job:s.
	*/
	public Collection<Permissions> getPermissions()
	{
		if (permissions.size() == 0)
		{
			permissions.add(new Permissions(Item.EXPERIMENT, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.EXTRAVALUETYPE, null, EnumSet.of(Permission.USE)));
			permissions.add(new Permissions(Item.JOB, null, EnumSet.of(Permission.READ)));
		}
		return permissions;
	}
	public void run(Request request, Response response, ProgressReporter progress)
	{
		if (signalHandler != null) signalHandler.setWorkerThread(null);
		DbControl dc = sc.newDbControl();
		try
		{
			if (progress != null) progress.display(5, "Initialising expressions...");
			
			// Get all parameters
			BioAssaySet source = getSourceBioAssaySet(dc);
			List<BioAssay> bioAssays = getSourceBioAssays(dc);
			ExtraValueType extraValueType = (ExtraValueType)job.getValue("extraValueType");
			String expression = (String)job.getValue("expression");
			
			// Get the current job
			Job currentJob = getCurrentJob(dc);
			
			// Create a batcher
			SpotExtraValueBatcher<?> extraBatcher = 
				source.getSpotExtraValueBatcher(extraValueType.getValueType().getValueClass(), 
					extraValueType, currentJob);
			
			// Create the query for source data
			DynamicSpotQuery query = source.getSpotData();
			query.setDisableRawJoinOnMergedData(true);
			query.select(Dynamic.select(VirtualColumn.COLUMN));
			query.select(Dynamic.select(VirtualColumn.POSITION));
			Expression extraExpression = BioAssaySetUtil.createJepExpression(dc, expression, query);
			query.select(Selects.expression(extraExpression, "xtra"));
			// Restrict to selected bioassays if there are any
			query = restrictSource(query, bioAssays);

			if (progress != null) progress.display(10, "Calculating extra value ("+source.getNumSpots()+" total)...");
			ThreadSignalHandler.checkInterrupted();
			extraBatcher.insert(query);

			if (progress != null) progress.display(90, "Committing transaction ("+source.getNumSpots()+" done)...");
			dc.commit();

			if (progress != null) progress.display(100, "Done");
			response.setDone("Calculated extra values for " + source.getNumSpots() + " spots.");
		}
		catch (Throwable t)
		{
			response.setError(t.getMessage(), Arrays.asList(t));
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	// -------------------------------------------

	/*
		From the InteractivePlugin interface
		-------------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	@Override
	public String isInContext(GuiContext context, Object item)
	{
		String message = super.isInContext(context, item);
		if (message == null)
		{
			BioAssaySet bas = null;
			if (context.equals(CONTEXT_BIOASSAYSET))
			{
				bas = (BioAssaySet)item;
			}
			else
			{
				DbControl dc = sc.newDbControl();
				try
				{
					bas = getCurrentBioAssaySet(dc);
				}
				finally
				{
					if (dc != null) dc.close();
				}
			}
			if (bas.getNumSpots() <= 0)
			{
				message = "This plug-in requires spot data in the database";
			}
		}
		return message;
	}
	public RequestInformation getRequestInformation(GuiContext context, String command) 
		throws BaseException
	{
		RequestInformation requestInformation = null;
		if (command.equals(Request.COMMAND_CONFIGURE_JOB))
		{
			requestInformation = getConfigureJobParameters();
		}
		return requestInformation;
	}

	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				// Validate job parameters
				List<Throwable> errors = 
					validateRequestParameters(getConfigureJobParameters().getParameters(), request);
				if (errors != null)
				{
					response.setError(errors.size() + " invalid parameter(s) were found in the request", errors);
					return;
				}
				
				// Source bioassay set
				storeValue(job, request, getSourceBioAssaySetParameter(null, null));
				
				// Subset of bioassays
				storeValue(job, request, this.getSourceBioAssaysParameter(null, null));
				
				// ExtraValueType
				storeValue(job, request, extraValueTypeParameter);
				
				// Expression
				storeValue(job, request, expressionParameter);
				
				BioAssaySet bioAssaySet = (BioAssaySet)job.getValue(SOURCE_BIOASSAYSET);
				ExtraValueType extraValueType = (ExtraValueType)job.getValue("extraValueType");
				response.setSuggestedJobName("Create extra value '" + extraValueType.getName() + 
						"' for bioassay set '" + bioAssaySet.getName() + "': " + job.getValue("expression"));
				
				response.setDone("Job configuration complete", Job.ExecutionTime.SHORT);
			}
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// -------------------------------------------

	/*
		From the SignalTarget interface
		-------------------------------------------
	*/
	public SignalHandler getSignalHandler()
	{
		signalHandler = new ThreadSignalHandler();
		return signalHandler;
	}
	// -------------------------------------------

	private RequestInformation getConfigureJobParameters()
	{
		DbControl dc = null;
		try
		{
			if (configureJob == null)
			{
				dc = sc.newDbControl();
				List<PluginParameter<?>> parameters = new ArrayList<PluginParameter<?>>(3);
				
				// Source bioassay set
				parameters.add(getSourceBioAssaySetParameter(null, null));
				
				// Sub set of bioassays
				parameters.add(getSourceBioAssaysParameter(null, null));
				
				// Extra value types
				ItemQuery<ExtraValueType> query = ExtraValueType.getQuery();
				query.order(Orders.asc(Hql.property("name")));
				ItemParameterType<ExtraValueType> extraValueTypeType = 
					new ItemParameterType<ExtraValueType>(ExtraValueType.class, null, true, 1, query.list(dc));
				
				extraValueTypeParameter = new PluginParameter<ExtraValueType>
				(
					"extraValueType", "Extra value type", "Select the type of the extra value",
					extraValueTypeType
				);
				
				parameters.add(extraValueTypeParameter);
				
				// Expression
				parameters.add(expressionParameter);

				configureJob = new RequestInformation
				(
					Request.COMMAND_CONFIGURE_JOB,
					"Specify extra value type and expression",
					"Select an extra value type and enter an expression to use for calculating the " +
					"extra values.",
					parameters, "jep_extra_value_calculator.jsp"
				);
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return configureJob;
	}

}
