/**
	$Id: ScanImporter.java 5371 2010-06-24 08:46:52Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.File;
import net.sf.basedb.core.Hardware;
import net.sf.basedb.core.Hybridization;
import net.sf.basedb.core.Image;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Path;
import net.sf.basedb.core.PathParameterType;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Protocol;
import net.sf.basedb.core.ProtocolType;
import net.sf.basedb.core.Scan;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Version;
import net.sf.basedb.core.plugin.About;
import net.sf.basedb.core.plugin.AboutImpl;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.Mapper;

import java.util.Collections;
import java.util.List;
import java.util.Set;

/**
	Plug-in for importing scan items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@version 2.8
	@base.modified $Date: 2010-06-24 10:46:52 +0200 (Thu, 24 Jun 2010) $
*/
public class ScanImporter
	extends AbstractItemImporter<Scan>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.SCAN, GuiContext.Type.LIST));

	private static final About about =
		new AboutImpl
	(
			"Scan importer",
			"Imports and updates scans in a batch.",
			Version.getMajor() + "." + Version.getMinor() + "." + Version.getMaintenance(),
			"2008, Base 2 development team",
			null,
			null,
			"http://base.thep.lu.se"
	);

	
	protected static final PluginParameter<String> hardwareColumnMapping = new PluginParameter<String>(
			"hardwareColumnMapping",
			"Scanner",
			"Mapping that picks the name or ID of the array slide scanner from the data columns. " +
			"The plug-in will first try to find a scanner with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Scanner\\",
			optionalColumnMapping
			);

	protected static final PluginParameter<String> hybridizationColumnMapping = new PluginParameter<String>(
			"hybridizationColumnMapping",
			"Hybridization",
			"Mapping that picks the name or ID of the hybridization from the data columns. " +
			"The plug-in will first try to find a hybridization with the given name. If none is found and " +
			"the value is numeric it will try to load by internal ID. " +
			"Example: \\Hybridization\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> imageFileColumnMapping = new PluginParameter<String>(
			"imageFileColumnMapping",
			"Image file",
			"Mapping that picks the path of a image file from the columns. " +
			"The path can be an absolute path (starting with /) or relative " +
			"to the 'Data directory' parameter." +
			"The plug-in tries to find an existing image with the given file path and that belongs to current scan. "+
			"Example: \\Image file\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> imageNameColumnMapping = new PluginParameter<String>(
			"imageColumnMapping",
			"Image name",
			"Mapping that picks the name the image from the data columns. " +			
			"Example: \\Image\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> imageFormatColumnMapping = new PluginParameter<String>(
			"imageFormatColumnMapping",
			"Image format",
			"Mapping that picks the format of the given image from the data columns. " +
			"The plug-in tries to parse the given format into one of the supported in BASE. " +
			"The image will get format 'unknown' if this isn't specified or has null-value." +
			"The column can either be empty or has one of the values 'TIFF', 'JPEG' or 'Unknown'."+
			"Example: \\Image format\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> imagePreviewColumnMapping = new PluginParameter<String>(
			"imagePreviewColumnMapping",
			"Image preview",
			"Mapping that picks the preview flag of the given image from the data columns. " +
			"Example: \\Image preview\\",
			optionalColumnMapping
			);
	
	protected static final PluginParameter<String> imageDescriptionColumnMapping = new PluginParameter<String>(
			"imageDescriptionColumnMapping",
			"Image description",
			"Mapping that picks the description of the given image from the data columns. " +
			"Example: \\Image description\\",
			optionalColumnMapping
			);

	
	public ScanImporter()
	{}

	/*
		From the Plugin interface
		------------------------------------
	*/
	@Override
	public About getAbout()
	{
		return about;
	}
	// ------------------------------------

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper descriptionMapper;
	private Mapper protocolMapper;
	private Mapper hardwareMapper;
	private Mapper hybridizationMapper;
	private Mapper imageFileMapper;
	private Mapper imageNameMapper;
	private Mapper imageFormatMapper;
	private Mapper imagePreviewMapper;
	private Mapper imageDescriptionMapper;
	private String imagesDirectory;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME,  
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	/**
		Adds column mappings for name, internalId, description,
		hybridization, protocol, hardware, image-file, image-name,
		image-format and image-description.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(hybridizationColumnMapping);
		parameters.add(protocolColumnMapping);
		parameters.add(hardwareColumnMapping);
		parameters.add(imageFileColumnMapping);
		parameters.add(imageNameColumnMapping);
		parameters.add(imageFormatColumnMapping);
		parameters.add(imagePreviewColumnMapping);
		parameters.add(imageDescriptionColumnMapping);
		parameters.add(permissionTemplateColumnMapping);
	}
	
	/**
		Add parameter for "Image directory" directory where image files may be located. 
	*/
	@Override
	protected void addMoreJobParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(new PluginParameter<String>(
				"imagesDirectory",
				"Images directory",
				"A base directory were image files are located. Relative file paths are " +
				"resolved against this directory. If no directory is given, only absolute " +
				"paths (=paths starting with /) can be used.",
				new PathParameterType(Path.Type.DIRECTORY))
		);
	}

	
	/**
		Calls {@link Scan#getQuery()}.
	*/
	@Override
	protected ItemQuery<Scan> createItemQuery()
	{
		return Scan.getQuery();
	}
	
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? Scan.MAX_NAME_LENGTH : null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? Scan.MAX_DESCRIPTION_LENGTH : null, null);
		protocolMapper = getMapper(ffp, (String)job.getValue("protocolColumnMapping"), null, null);
		hardwareMapper = getMapper(ffp, (String)job.getValue("hardwareColumnMapping"), null, null);
		hybridizationMapper = getMapper(ffp, (String)job.getValue("hybridizationColumnMapping"), null, null);
		imageFileMapper = getMapper(ffp, (String)job.getValue("imageFileColumnMapping"), null, null);
		imageNameMapper = getMapper(ffp, (String)job.getValue("imageColumnMapping"), null, null);
		imageFormatMapper = getMapper(ffp, (String)job.getValue("imageFormatColumnMapping"), null, null);
		imagePreviewMapper = getMapper(ffp, (String)job.getValue("imagePreviewColumnMapping"), null, null);
		imageDescriptionMapper = getMapper(ffp, (String)job.getValue("imageDescriptionColumnMapping"), null, null);
		imagesDirectory = (String)job.getValue("imagesDirectory");
		if (imagesDirectory != null && !imagesDirectory.startsWith("/"))
		{
			imagesDirectory += "/";
		}
	}
	
	@Override
	protected Scan createItem(DbControl dc, FlatFileParser.Data data)
	{
		Hybridization hyb = null;
		if (hybridizationMapper != null) 
		{
			hyb = findHybridization(dc, FallbackIdMethod.NAME_OR_ID, hybridizationMapper.getValue(data));
		}

		Scan scan = Scan.getNew(dc, hyb);
		updateItem(dc, scan, data);
		return scan;
	}
	
	@Override
	protected void updateItem(DbControl dc, Scan scan, FlatFileParser.Data data)
	{
		if (nameMapper != null) scan.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) scan.setDescription(descriptionMapper.getValue(data));
		if (hybridizationMapper != null && scan.isInDatabase()) 
		{
			String nameOrId = hybridizationMapper.getValue(data);
			if (nameOrId != null)
			{
				Hybridization hyb = findHybridization(dc, FallbackIdMethod.NAME_OR_ID, hybridizationMapper.getValue(data));
				scan.setHybridization(hyb);
			}
		}
		if (protocolMapper != null) 
		{
			String nameOrId = protocolMapper.getValue(data);
			ProtocolType type = ProtocolType.getById(dc, SystemItems.getId(ProtocolType.SCANNING));
			Protocol protocol = findProtocol(dc, FallbackIdMethod.NAME_OR_EXTERNALID_OR_ID, nameOrId, type);
			if (nameOrId == null || protocol != null) scan.setProtocol(protocol);
		}
		if (hardwareMapper != null)
		{
			String nameOrId = hardwareMapper.getValue(data);
			Hardware scanner = findScanner(dc, FallbackIdMethod.NAME_OR_ID, nameOrId);
			if (nameOrId == null || scanner != null) scan.setScanner(scanner);
		}
		
		updateMultiLineItem(dc, scan, data, 0);
	}
	
	@Override
	protected void updateMultiLineItem(DbControl dc, Scan scan, FlatFileParser.Data data, int multiLineNum)
	{
		Image image = null;
		File imageFile = null;
		String imageName = null;
		// Get the imagefile
		if (imageFileMapper != null && imageFileMapper.getValue(data) != null)
		{
			String filePath = imageFileMapper.getValue(data);
			if (!filePath.startsWith("/") && imagesDirectory != null)
			{
				filePath = imagesDirectory + filePath;
			}
			imageFile = File.getByPath(dc, new Path(filePath, Path.Type.FILE), false);
			image = findImage(dc, scan, imageFile);
		}
		if (imageNameMapper != null) 
		{
			imageName = imageNameMapper.getValue(data);
		}
		else if (image != null)
		{
			imageName = image.getName();
		}
		
		if (image != null)
		{
			image.setName(imageName != null ? imageName : scan.getName() + "." + imageFile.getName());
			if (imageDescriptionMapper != null) image.setDescription(imageDescriptionMapper.getValue(data));
			if (imageFormatMapper != null && imageFormatMapper.getValue(data) != null) image.setFormat(Image.Format.valueOf(imageFormatMapper.getValue(data)));
			if (imagePreviewMapper != null) image.setPreview(Boolean.getBoolean(imagePreviewMapper.getValue(data)));
		}
	}
	// ------------------------------------
}
