/*
	$Id: TestExperiment.java 5340 2010-05-10 06:15:00Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg, Gregory Vincic
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/



import java.util.Date;
import java.util.Set;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Experiment;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemProxy;
import net.sf.basedb.core.ItemResultList;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Platform;
import net.sf.basedb.core.RawBioAssay;
import net.sf.basedb.core.RawDataTypes;
import net.sf.basedb.core.Type;

public class TestExperiment
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing experiments");
		write_header();

		// Standard tests: create, load, list
		int id = test_create("genepix", true);
		int id2 = test_create("genepix", false);
		
		test_load(id);
		test_list(-1);
		
		// Experimental factors
		int factorId1 = TestAnnotationType.test_create("Drug resistance", Type.STRING, 0, null, null, Item.RAWBIOASSAY, 1, new String[] { "high", "medium", "low" }, false);
		int factorId2 = TestAnnotationType.test_create("Temperature", Type.FLOAT, 0, null, null, Item.RAWBIOASSAY, 1, null, false);
		test_add_experimental_factors(id, factorId1, factorId2);
		test_list_experimental_factors(id, 2);
		
		// Import reporters
		TestReporter.test_import_from_file("data/test.rawdata.import.txt", 
			"\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", "\\t", 4, 3);
		
		int arrayDesignId = TestArrayDesign.test_create(Platform.GENERIC, false);
		TestArrayDesign.test_import_from_file(arrayDesignId, "data/test.rawdata.import.txt", 
			"\"Block\"\\t\"Column\"\\t\"Row\"\\t\"Name\"\\t\"ID\".*", null, 0, 2, 1, 4);
		
		int rbaId1 = TestRawBioAssay.test_create(Platform.GENERIC, "genepix", null, 0, 0, 0, arrayDesignId, 1, false);
		int rbaId2 = TestRawBioAssay.test_create(Platform.GENERIC, "genepix", null, 0, 0, 0, arrayDesignId, 1, false);
		TestRawBioAssay.test_import_from_file(rbaId1, "data/test.rawdata.import.txt");
		TestRawBioAssay.test_import_from_file(rbaId2, "data/test.rawdata.import.txt");

		test_add_rawbioassay(id, rbaId1);
		test_add_rawbioassay(id, rbaId2);
		test_list_rawbioassays(id, 2);

		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		
		test_remove_rawbioassay(id, rbaId1);
		test_remove_rawbioassay(id, rbaId2);

		// Standard test: Delete
		test_delete(id);
		test_delete(id2);

		TestAnnotationType.test_delete(factorId1);
		TestAnnotationType.test_delete(factorId2);
		TestRawBioAssay.test_delete(rbaId1);
		TestRawBioAssay.test_delete(rbaId2);
		TestArrayDesign.test_delete(arrayDesignId);
		TestReporter.test_delete();
		write("++Testing experiments "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(String rawDataType, boolean setAll)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.EXPERIMENT)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getNew(dc, RawDataTypes.getSafeRawDataType(rawDataType));
			if (setAll)
			{
				e.setName("Test experiment");
				e.setDescription("Added at "+new Date());
				e.setAbstract("An abstract");
				e.setAffiliations("The affiliations");
				e.setAuthors("The authors");
				e.setExperimentDesign("A cool design");
				e.setExperimentType("The type");
				e.setPublication("Here is the publication");
				e.setPublicationDate(new Date());
				e.setPubMedId("PUBMED-123ABC");
			}
			dc.saveItem(e);
			dc.commit();
			id = e.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(e, false);
			write_item(0, e);
			write("--Create experiment OK");
		}
		catch (Throwable ex)
		{
			write("--Create experiment FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);
			write_item(0, e);
			write("--Load experiment OK");
		}
		catch (Throwable ex)
		{
			write("--Load experiment FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Experiment> l = Experiment.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List experiments OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List experiments FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);
			dc.deleteItem(e);
			Set<ItemProxy> using = e.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(e + " is used by " + using);
			}
			dc.commit();
			write("--Delete experiment OK");
		}
		catch (Throwable ex)
		{
			write("--Delete experiment FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tRaw data type\tDescription\tAbstract\tAffiliations\tAuthors\tDesign\tType\tPublication\tDate\tPubMed ID");
			write("-- \t-- \t--------- \t-------------\t-----------\t--------\t------------\t-------\t------\t----\t-----------\t----\t---------");
		}
	}
	static void write_item(int i, Experiment e)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+e.getId()+"\t"+e.getName()+"\t"+e.getRawDataType().getName()+"\t"+e.getDescription()+
				"\t"+e.getAbstract()+"\t"+e.getAffiliations()+"\t"+e.getAuthors()+
				"\t"+e.getExperimentDesign()+"\t"+e.getExperimentType()+"\t"+e.getPublication()+
				"\t"+e.getPublicationDate()+"\t"+e.getPubMedId());
		}
	}
	static void write_item(int i, RawBioAssay rba)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			System.out.println(i+":\t"+rba.getId()+"\t"+rba.getName()+"\t"+rba.getDescription()+
				"\t"+rba.getRawDataType().getName()+"\t"+rba.getNumDbSpots()+"\t"+rba.getBytes()+"\t"+rba.getArrayDesign()+
				"\t"+rba.getSoftware()+"\t"+rba.getProtocol());
		}
	}
	
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void test_add_rawbioassay(int id, int... rawBioAssayId)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);
			for (int rbaId : rawBioAssayId)
			{
				RawBioAssay rba = RawBioAssay.getById(dc, rbaId);
				e.addRawBioAssay(rba);
			}
			dc.commit();
			write("--Add raw bioassays OK");
		}
		catch (Throwable ex)
		{
			write("--Add raw bioassays FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_remove_rawbioassay(int id, int rawBioAssayId)
	{
		if (id == 0 || rawBioAssayId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);
			RawBioAssay rba = RawBioAssay.getById(dc, rawBioAssayId);
			e.removeRawBioAssay(rba);
			dc.commit();
			write("--Remove raw bioassay OK");
		}
		catch (Throwable ex)
		{
			write("--Remove raw bioassay FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static int[] test_list_rawbioassays(int id, int expectedResults)
	{
		if (id == 0) return null;
		DbControl dc = null;
		int[] rawBioAssays = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);

			ItemResultList<RawBioAssay> l = e.getRawBioAssays().list(dc);
			rawBioAssays = new int[l.size()];
			for (int i = 0; i<l.size(); i++)
			{
				rawBioAssays[i] = l.get(i).getId();
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List raw bioassays OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List raw bioassays FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return rawBioAssays;
	}
	
	static void test_add_experimental_factors(int id, int... annotationTypes)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);
			for (int atId : annotationTypes)
			{
				AnnotationType at = AnnotationType.getById(dc, atId);
				e.addExperimentalFactor(at);
			}
			dc.commit();
			write("--Add experimental factors OK");
		}
		catch (Throwable ex)
		{
			write("--Add experimental factors FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_remove_experimental_factors(int id, int... annotationTypes)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);
			for (int atId : annotationTypes)
			{
				AnnotationType at = AnnotationType.getById(dc, atId);
				e.removeExperimentalFactor(at);
			}
			dc.commit();
			write("--Remove experimental factors OK");
		}
		catch (Throwable ex)
		{
			write("--Remove experimental factors FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	static void test_list_experimental_factors(int id, int expectedResults)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Experiment e = Experiment.getById(dc, id);

			ItemResultList<AnnotationType> l = e.getExperimentalFactors().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				TestAnnotationType.write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List experimental factors OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List experimental factors FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}	

}
