/*
	$Id: TestHybridization.java 5340 2010-05-10 06:15:00Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
import net.sf.basedb.core.*;

import java.util.Date;
import java.util.Set;

public class TestHybridization
{

	static boolean ok = true;
	public static void main(String[] args)
	{
		TestUtil.checkArgs(args);
		TestUtil.begin();
		ok = test_all();
		TestUtil.stop();
	}

	static boolean test_all()
	{
		write("++Testing hybridizations");
		write_header();
		// Standard tests: create, load, list
		int id = test_create(0, true);
		int id2 = test_create(0, false);
		
		test_load(id);

		int arrayDesignId = TestArrayDesign.test_create(Platform.GENERIC, false);
		int arrayBatchId = TestArrayBatch.test_create(arrayDesignId, false);
		int arraySlideId = TestArraySlide.test_create(arrayBatchId, false);
		int id3 = test_create(arraySlideId, false);
		
		int labelId1 = TestLabel.test_create(false);
		int labelId2 = TestLabel.test_create(false);
		int labeledExtractId1 = TestLabeledExtract.test_create(0, labelId1, false);
		int labeledExtractId2 = TestLabeledExtract.test_create(0, labelId2, false);
		int id4 = test_create(0, false, labeledExtractId1, labeledExtractId2);
		
		test_list(-1);
		
		TestLabeledExtract.test_list_hybridizations(labeledExtractId1, 1);
		ok = ok && TestLabeledExtract.ok;

		// Standard test: Delete
		if (TestUtil.waitBeforeDelete()) TestUtil.waitForEnter();
		test_delete(id);
		test_delete(id2);
		test_delete(id3);
		test_delete(id4);
		
		TestLabeledExtract.test_delete(labeledExtractId1);
		TestLabeledExtract.test_delete(labeledExtractId2);
		TestLabel.test_delete(labelId1);
		TestLabel.test_delete(labelId2);
		
		TestArraySlide.test_delete(arraySlideId);
		TestArrayBatch.test_delete(arrayBatchId);
		TestArrayDesign.test_delete(arrayDesignId);
		write("++Testing hybridizations "+(ok ? "OK" : "Failed")+"\n");
		return ok;
	}

	static int test_create(int arraySlideId, boolean setAll, int... labeledExtractIds)
	{
		if (!TestUtil.hasPermission(Permission.CREATE, Item.HYBRIDIZATION)) return 0;
		int id = 0;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Hybridization h = Hybridization.getNew(dc);
			if (setAll)
			{
				h.setName("Test hybridization");
				h.setDescription("Added at "+new Date());
			}
			if (arraySlideId != 0)
			{
				h.setArraySlide(ArraySlide.getById(dc, arraySlideId));
			}
			BioMaterialEvent evt = h.getCreationEvent();
			for (int labeledExtractId : labeledExtractIds)
			{
				LabeledExtract le = LabeledExtract.getById(dc, labeledExtractId);
				evt.addSource(le, 100.0f);
			}
			dc.saveItem(h);
			dc.commit();
			id = h.getId();
			dc = TestUtil.getDbControl();
			dc.reattachItem(h, false);
			dc.reattachItem(h.getCreationEvent(), false);
			write_item(0, h);
			write("--Create hybridization OK");
		}
		catch (Throwable ex)
		{
			write("--Create hybridization FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
		return id;
	}

	static void test_load(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Hybridization h = Hybridization.getById(dc, id);
			write_item(0, h);
			write("--Load hybridization OK");
		}
		catch (Throwable ex)
		{
			write("--Load hybridization FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_list(int expectedResults)
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Hybridization> l = Hybridization.getQuery().list(dc);
			for (int i = 0; i<l.size(); i++)
			{
				write_item(i, l.get(i));
			}
			if (expectedResults >= 0 && expectedResults != l.size())
			{
				throw new BaseException("Expected "+expectedResults+" results, not "+l.size());
			}
			write("--List hybridizations OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--List hybridizations FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete(int id)
	{
		if (id == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Hybridization h = Hybridization.getById(dc, id);
			dc.deleteItem(h);
			Set<ItemProxy> using = h.getUsingItems();
			if (using.size() > 0) 
			{
			   throw new BaseException(h + " is used by " + using);
			}
			dc.commit();
			write("--Delete hybridization OK");
		}
		catch (Throwable ex)
		{
			write("--Delete hybridization FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	static void test_delete_all()
	{
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			ItemResultList<Hybridization> l = Hybridization.getQuery().list(dc);
			Trashcan.delete(dc.getSessionControl(), l, true, null);
			dc.commit();
			write("--Delete all hybridizations OK ("+l.size()+")");
		}
		catch (Throwable ex)
		{
			write("--Delete all hybridizations FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	
	static void write_header()
	{
		if (!TestUtil.getSilent())
		{
			write("   \tID \tName      \tDescription\tArray slide\tLabeled extracts");
			write("-- \t-- \t--------- \t-----------\t-----------\t----------------");
		}
	}
	static void write_item(int i, Hybridization h)
		throws BaseException
	{
		if (!TestUtil.getSilent()) 
		{
			StringBuilder sb = new StringBuilder();
			sb.append("[");
			for (MeasuredBioMaterial le : h.getCreationEvent().getSources().list(h.getDbControl()))
			{
				sb.append(le.toString());
				sb.append(",");
			}
			sb.append("]");
			System.out.println(i+":\t"+h.getId()+"\t"+h.getName()+"\t"+h.getDescription()+
				"\t"+h.getArraySlide()+"\t"+sb);
		}
	}
	static void write(String message)
	{
		System.out.println(message);
	}
	
	static void test_add_labeledextract(int hybId, int labeledExtractId, int arrayNum, Float usedQuantity)
	{
		if (hybId == 0 || labeledExtractId == 0) return;
		DbControl dc = null;
		try
		{
			dc = TestUtil.getDbControl();
			Hybridization h = Hybridization.getById(dc, hybId);
			LabeledExtract le = LabeledExtract.getById(dc, labeledExtractId);
			BioMaterialEvent event = h.getCreationEvent();
			event.addSource(le, usedQuantity);
			event.setSourceGroup(le, arrayNum);
			dc.commit();
			write("--Add labeled extract to hybridization OK");
		}
		catch (Throwable ex)
		{
			write("--Add labeled extract to hybridization FAILED");
			ex.printStackTrace();
			ok = false;
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
}
